# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import json
from typing import Any, Optional, Type

import pydantic
import pydantic_core

from ..collection_utils import get_by_path
from .http import HttpError

AUTHENTICATION_FAILURE_MESSAGE = (
    "Authentication to Roboto failed. This generally happens if the value in your "
    + "ROBOTO_BEARER_TOKEN environment variable or token in your ~/.roboto/config.json "
    + "file is expired or incorrectly formatted."
)


class RobotoDomainException(Exception):
    """
    Expected exceptions from the Roboto domain entity objects.
    """

    __headers: dict[str, str]
    _message: str
    _stack_trace: list[str]

    def __init__(
        self,
        message: str,
        stack_trace: list[str] = [],
        headers: dict[str, str] = {},
        *args,
        **kwargs,
    ):
        super().__init__(message, *args, **kwargs)
        self._message = message
        self._stack_trace = stack_trace
        self.__headers = headers

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source: Type[Any], handler: pydantic.GetCoreSchemaHandler
    ) -> pydantic_core.core_schema.CoreSchema:
        assert source is RobotoDomainException
        return pydantic_core.core_schema.no_info_after_validator_function(
            cls.from_json_string,
            pydantic_core.core_schema.str_schema(),
            serialization=pydantic_core.core_schema.plain_serializer_function_ser_schema(
                cls.to_dict,
                info_arg=False,
                return_schema=pydantic_core.core_schema.str_schema(),
            ),
        )

    @staticmethod
    def from_json(
        contents: dict[str, Any], headers: dict[str, str] = {}
    ) -> "RobotoDomainException":
        error_code = get_by_path(contents, ["error", "error_code"])
        inner_message = get_by_path(contents, ["error", "message"])
        kwargs: dict[str, Any] = {}
        error = get_by_path(contents, ["error"])
        if error is not None:
            kwargs.update(error)
            kwargs.pop("error_code")
            kwargs.pop("message")

        if error_code is None or inner_message is None:
            raise ValueError("Need 'error_code' and 'message' available.")

        for subclass in RobotoDomainException.__subclasses__():
            if subclass.__name__ == error_code:
                return subclass(message=inner_message, headers=headers, **kwargs)

        raise ValueError("Unrecognized error code 'error_code'")

    @staticmethod
    def from_json_string(contents: str) -> "RobotoDomainException":
        return RobotoDomainException.from_json(json.loads(contents))

    @staticmethod
    def from_client_error(error: HttpError) -> "RobotoDomainException":
        message: Optional[str]

        if type(error.msg) is dict:
            # See if it's a first class RobotoException
            try:
                return RobotoDomainException.from_json(error.msg, headers=error.headers)
            except ValueError:
                pass

            # Handle JSON from non-roboto calls
            message = error.msg.get("message", json.dumps(error.msg))
        elif type(error.msg) is str:
            message = error.msg
        else:
            message = None

        if error.status is None:
            raise RobotoDomainException(error.msg, headers=error.headers)
        if error.status == 400:
            if (
                message is not None
                and "did not provide a org for single-org operation" in message
            ):
                return RobotoNoOrgProvidedException(error.msg, headers=error.headers)
            else:
                return RobotoInvalidRequestException(error.msg, headers=error.headers)
        if error.status in (401, 403):
            if (
                message is not None
                and "User is not authorized to access this resource with an explicit deny"
                in message
            ):
                return RobotoAuthenticationFailureException(
                    AUTHENTICATION_FAILURE_MESSAGE, headers=error.headers
                )
            else:
                return RobotoUnauthorizedException(error.msg, headers=error.headers)
        if error.status == 404:
            return RobotoNotFoundException(error.msg, headers=error.headers)
        if 500 <= error.status < 600:
            return RobotoServiceException(error.msg, headers=error.headers)
        raise error

    @property
    def headers(self) -> dict[str, str]:
        return self.__headers

    @property
    def http_status_code(self) -> int:
        return 500

    @property
    def error_code(self) -> str:
        return self.__class__.__name__

    @property
    def message(self) -> str:
        return self._message

    @property
    def stack_trace(self) -> list[str]:
        return self._stack_trace

    @stack_trace.setter
    def stack_trace(self, stack_trace: list[str]):
        self._stack_trace = stack_trace

    def to_dict(self) -> dict[str, Any]:
        error: dict[str, Any] = {"error_code": self.error_code, "message": self.message}
        if len(self._stack_trace) > 0:
            error["stack_trace"] = self._stack_trace
        return {"error": error}

    def serialize(self) -> str:
        return json.dumps(self.to_dict())


class RobotoUnauthorizedException(RobotoDomainException):
    """
    Thrown when a user is attempting to access a resource that they do not have permission to access
    """

    @property
    def http_status_code(self) -> int:
        return 401


class RobotoAuthenticationFailureException(RobotoDomainException):
    """
    Thrown when authentication fails
    """

    @property
    def http_status_code(self) -> int:
        return 401


class RobotoNotFoundException(RobotoDomainException):
    """
    Throw when a requested resource does not exist
    """

    @property
    def http_status_code(self) -> int:
        return 404


class RobotoNotReadyException(RobotoDomainException):
    """
    Throw when a requested resource is resolvable by ID/context but is not fully initialized or ready to be served.
    """

    @property
    def http_status_code(self) -> int:
        return 404


class RobotoIllegalArgumentException(RobotoDomainException):
    """
    Thrown when request parameters are in some way invalid
    """

    @property
    def http_status_code(self) -> int:
        return 400


class RobotoInvalidRequestException(RobotoDomainException):
    """
    Thrown when request parameters are in some way invalid
    """

    @property
    def http_status_code(self) -> int:
        return 400


class RobotoNotImplementedException(RobotoDomainException):
    """
    Thrown by shimmed out APIs which have not yet been implemented
    """

    @property
    def http_status_code(self) -> int:
        return 501


class RobotoDeprecatedException(RobotoDomainException):
    """
    Thrown when an old API endpoint is called, to tell the client to upgrade
    """

    @property
    def http_status_code(self) -> int:
        return 400


class RobotoInvalidStateTransitionException(RobotoDomainException):
    """
    Thrown when requesting update of state to an invalid state, or via an invalid transition path
    """

    @property
    def http_status_code(self) -> int:
        return 400


class RobotoNoOrgProvidedException(RobotoDomainException):
    """
    Thrown when no org is provided to an operation which requires an org.
    """

    @property
    def http_status_code(self) -> int:
        return 400


class RobotoConditionException(RobotoDomainException):
    """
    Thrown if there is a failed condition
    """

    @property
    def http_status_code(self) -> int:
        return 409


class RobotoConflictException(RobotoDomainException):
    """
    Thrown if there is a conflict between a resource you're creating and another existing resource
    """

    @property
    def http_status_code(self) -> int:
        return 409


class RobotoServiceException(RobotoDomainException):
    """
    Thrown when Roboto Service failed in an unexpected way
    """


class RobotoUnknownOperationException(RobotoDomainException):
    """
    Thrown if a user is attempting to perform an action unrecognized by the Roboto platform.
    """

    @property
    def http_status_code(self) -> int:
        return 404


class RobotoExpiredException(RobotoDomainException):
    """
    Thrown if a resource is missing or expired.
    """

    @property
    def http_status_code(self) -> int:
        return 410


class RobotoLimitExceededException(RobotoDomainException):
    """
    Thrown if an operation would exceed a user or org level limit.
    """

    __resource_name: str
    __current_quantity: int
    __limit_quantity: int

    def __init__(
        self,
        message: str,
        stack_trace: list[str] = [],
        *args,
        resource_name: str = "Unknown",
        current_quantity: int = 0,
        limit_quantity: int = 0,
        **kwargs,
    ):
        super().__init__(message, stack_trace, *args, **kwargs)
        self.__resource_name = resource_name
        self.__current_quantity = current_quantity
        self.__limit_quantity = limit_quantity

    @property
    def http_status_code(self) -> int:
        return 403

    @property
    def resource_name(self) -> str:
        return self.__resource_name

    @property
    def limit_quantity(self) -> int:
        return self.__limit_quantity

    @property
    def current_quantity(self) -> int:
        return self.__current_quantity

    def to_dict(self) -> dict[str, Any]:
        as_dict = super().to_dict()
        as_dict["error"]["resource_name"] = self.resource_name
        as_dict["error"]["current_quantity"] = self.current_quantity
        as_dict["error"]["limit_quantity"] = self.limit_quantity
        return as_dict


class RobotoInternalException(RobotoDomainException):
    """
    Thrown when Roboto throws an unexpected internal error, expected to be systemic.
    """

    @property
    def http_status_code(self) -> int:
        return 500


class RobotoServiceUnavailableException(RobotoDomainException):
    """
    Thrown when a service is unavailable, such as when it's under heavy load and can't accept new requests.
    This is expected to be transient and ought to be retried.
    """

    @property
    def http_status_code(self) -> int:
        return 503


class RobotoServiceTimeoutException(RobotoDomainException):
    """
    Thrown when the service times out while processing a request. This is exepcted to be transient and ought to be
    retried.
    """

    @property
    def http_status_code(self) -> int:
        return 504


class RobotoHttpExceptionParse(object):
    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception, traceback):
        if issubclass(type(exception), HttpError):
            raise RobotoDomainException.from_client_error(error=exception) from None



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

from __future__ import annotations

import collections.abc
import datetime
import pathlib
import typing

from ...association import Association
from ...compat import import_optional_dependency
from ...exceptions import (
    RobotoInvalidRequestException,
)
from ...http import RobotoClient
from ...logging import default_logger
from ...sentinels import (
    NotSet,
    NotSetType,
    remove_not_set,
)
from ...time import to_epoch_nanoseconds
from ...updates import (
    MetadataChangeset,
    StrSequence,
)
from ..files import File
from ..topics import (
    MessagePath,
    Topic,
    TopicDataService,
)
from .operations import (
    CreateEventRequest,
    QueryEventsForAssociationsRequest,
    UpdateEventRequest,
)
from .record import EventRecord

if typing.TYPE_CHECKING:
    import pandas  # pants: no-infer-dep

logger = default_logger()


class Event:
    """
    An event is a "time anchor" which allows you to relate first class Roboto entities
    (datasets, files, topics and message paths), as well as a timespan in which they occurred.
    """

    __roboto_client: RobotoClient
    __record: EventRecord

    @classmethod
    def create(
        cls,
        start_time: typing.Union[int, datetime.datetime],
        name: str = "Custom Event",
        associations: typing.Optional[collections.abc.Collection[Association]] = None,
        file_ids: typing.Optional[collections.abc.Collection[str]] = None,
        topic_ids: typing.Optional[collections.abc.Collection[str]] = None,
        dataset_ids: typing.Optional[collections.abc.Collection[str]] = None,
        message_path_ids: typing.Optional[collections.abc.Collection[str]] = None,
        end_time: typing.Optional[typing.Union[int, datetime.datetime]] = None,
        description: typing.Optional[str] = None,
        metadata: typing.Optional[dict[str, typing.Any]] = None,
        tags: typing.Optional[list[str]] = None,
        caller_org_id: typing.Optional[str] = None,
        roboto_client: typing.Optional[RobotoClient] = None,
    ) -> "Event":
        roboto_client = RobotoClient.defaulted(roboto_client)

        coalesced_associations = Association.coalesce(
            associations=associations,
            dataset_ids=dataset_ids,
            file_ids=file_ids,
            topic_ids=topic_ids,
            message_path_ids=message_path_ids,
            throw_on_empty=True,
        )

        request = CreateEventRequest(
            associations=coalesced_associations,
            start_time=to_epoch_nanoseconds(start_time),
            end_time=to_epoch_nanoseconds(end_time or start_time),
            description=description,
            metadata=metadata or {},
            tags=tags or [],
            name=name,
        )
        record = roboto_client.post(
            "v1/events/create", caller_org_id=caller_org_id, data=request
        ).to_record(EventRecord)
        return cls(record=record, roboto_client=roboto_client)

    @classmethod
    def get_by_dataset(
        cls,
        dataset_id: str,
        roboto_client: typing.Optional[RobotoClient] = None,
        transitive: bool = False,
    ) -> collections.abc.Generator["Event", None, None]:
        roboto_client = RobotoClient.defaulted(roboto_client)

        # This will return only events that explicitly have an association with this dataset
        if transitive is False:
            for event in Event.get_by_associations(
                [Association.dataset(dataset_id)], roboto_client
            ):
                yield event
            return

        # This will return all events that have an association with this dataset or any of its files or topics
        next_token: typing.Optional[str] = None
        while True:
            results = roboto_client.get(
                f"v1/datasets/{dataset_id}/events",
                query={"page_token": next_token} if next_token is not None else None,
            ).to_paginated_list(EventRecord)

            for item in results.items:
                yield cls(record=item, roboto_client=roboto_client)

            next_token = results.next_token
            if not next_token:
                break

    @classmethod
    def get_by_file(
        cls, file_id: str, roboto_client: typing.Optional[RobotoClient] = None
    ) -> collections.abc.Generator["Event", None, None]:
        """Returns all events with a direct association to the provided file."""
        return Event.get_by_associations([Association.file(file_id)], roboto_client)

    @classmethod
    def get_by_message_path(
        cls,
        message_path_id: str,
        roboto_client: typing.Optional[RobotoClient] = None,
    ) -> collections.abc.Generator["Event", None, None]:
        """Returns all events with a direct association to the provided message path."""
        return Event.get_by_associations(
            [Association.msgpath(message_path_id)], roboto_client
        )

    @classmethod
    def get_by_topic(
        cls,
        topic_id: str,
        roboto_client: typing.Optional[RobotoClient] = None,
    ) -> collections.abc.Generator["Event", None, None]:
        """Returns all events with a direct association to the provided topic."""
        return Event.get_by_associations([Association.topic(topic_id)], roboto_client)

    @classmethod
    def get_by_associations(
        cls,
        associations: collections.abc.Collection[Association],
        roboto_client: typing.Optional[RobotoClient] = None,
    ) -> collections.abc.Generator["Event", None, None]:
        """
        Returns all events associated with the provided association. Any events which you don't have access to will be
        filtered out of the response rather than throwing an exception.
        """
        roboto_client = RobotoClient.defaulted(roboto_client)

        next_token: typing.Optional[str] = None
        while True:
            request = QueryEventsForAssociationsRequest(
                associations=list(associations), page_token=next_token
            )

            results = roboto_client.post(
                "v1/events/query/for_associations",
                data=request,
            ).to_paginated_list(EventRecord)

            for item in results.items:
                yield cls(record=item, roboto_client=roboto_client)

            next_token = results.next_token
            if not next_token:
                break

    @classmethod
    def from_id(
        cls, event_id: str, roboto_client: typing.Optional[RobotoClient] = None
    ):
        roboto_client = RobotoClient.defaulted(roboto_client)
        record = roboto_client.get(f"v1/events/id/{event_id}").to_record(EventRecord)
        return cls(record, roboto_client)

    def __eq__(self, other: typing.Any) -> bool:
        if not isinstance(other, Event):
            return NotImplemented

        return self.__record == other.__record

    def __init__(
        self, record: EventRecord, roboto_client: typing.Optional[RobotoClient] = None
    ) -> None:
        self.__roboto_client = RobotoClient.defaulted(roboto_client)
        self.__record = record

    def __repr__(self) -> str:
        return self.__record.model_dump_json()

    @property
    def created(self) -> datetime.datetime:
        return self.__record.created

    @property
    def created_by(self) -> str:
        return self.__record.created_by

    @property
    def dataset_ids(self) -> list[str]:
        return [
            association.association_id
            for association in self.__record.associations
            if association.is_dataset
        ]

    @property
    def description(self) -> typing.Optional[str]:
        return self.__record.description

    @property
    def end_time(self) -> int:
        """Epoch nanoseconds"""
        return self.__record.end_time

    @property
    def event_id(self) -> str:
        return self.__record.event_id

    @property
    def file_ids(self) -> list[str]:
        return [
            association.association_id
            for association in self.__record.associations
            if association.is_file
        ]

    @property
    def message_path_ids(self) -> list[str]:
        return [
            association.association_id
            for association in self.__record.associations
            if association.is_msgpath
        ]

    @property
    def metadata(self) -> dict[str, typing.Any]:
        return self.__record.metadata

    @property
    def modified(self) -> datetime.datetime:
        return self.__record.modified

    @property
    def modified_by(self) -> str:
        return self.__record.modified_by

    @property
    def name(self) -> str:
        return self.__record.name

    @property
    def record(self) -> EventRecord:
        return self.__record

    @property
    def start_time(self) -> int:
        """Epoch nanoseconds"""
        return self.__record.start_time

    @property
    def tags(self) -> list[str]:
        return self.__record.tags

    @property
    def topic_ids(self) -> list[str]:
        return [
            association.association_id
            for association in self.__record.associations
            if association.is_topic
        ]

    def delete(self) -> None:
        self.__roboto_client.delete(f"v1/events/id/{self.event_id}")

    def get_data(
        self,
        message_paths_include: typing.Optional[collections.abc.Sequence[str]] = None,
        message_paths_exclude: typing.Optional[collections.abc.Sequence[str]] = None,
        topic_name: typing.Optional[str] = None,
        topic_data_service: typing.Optional[TopicDataService] = None,
        cache_dir: typing.Union[str, pathlib.Path, None] = None,
    ) -> collections.abc.Generator[dict[str, typing.Any], None, None]:
        """
        Iteratively yield records of the underlying topic data this event annotates.

        An event can be associated with data at multiple resolutions:
            - as an event on its containing dataset, file, and/or topic,
            - but also directly with the message path ("signal data")
        A single event can also span signals that share a timeline,
        so it may annotate multiple topics in a file, or even multiple files in a dataset.

        For now, getting the underlying signal data associated with an event only works
        for events that can be sourced to a single topic (extracted from one file, uploaded to one dataset).
        This means that the event must have been made on either a single file, a single topic,
        or one or many message paths within that topic.

        If the event was made on a file, `topic_name` must be provided,
        and either or both of `message_paths_include` or `message_paths_exclude` may be provided,
        but are optional.

        If the event was made on a topic, either or both of `message_paths_include` or `message_paths_exclude`
        may be provided, but are optional.
        `topic_name`, if provided in this instance, is ignored.

        If the event was made on one or many message paths,
        each of those message paths must be found in the same topic.
        `topic_name`, `message_paths_include`, and `message_paths_exclude`, if provided in this instance,
        are ignored.

        If the event is associated with data at multiple resolutions (e.g., two message paths, one topic, one file),
        this method will consider the lowest resolution associations first (message path), then topic, then file.

        If ``message_paths_include`` or ``message_paths_exclude`` are defined,
        they should be dot notation paths that match attributes of individual data records.
        If a partial path is provided, it is treated as a wildcard, matching all subpaths.

        For example, given topic data with the following interface:

        ::

            {
                "velocity": {
                    "x": <uint32>,
                    "y": <uint32>,
                    "z": <uint32>
                }
            }

        Calling ``get_data`` on an Event associated with that topic like:

            >>> event.get_data(message_paths_include=["velocity.x", "velocity.y"])

        is expected to give the same output as:

            >>> event.get_data(message_paths_include=["velocity"], message_paths_exclude=["velocity.z"])
        """
        # Event associated with message path(s)
        if len(self.message_path_ids) > 0:
            message_paths = [
                MessagePath.from_id(
                    message_path_id=message_path_id,
                    roboto_client=self.__roboto_client,
                    topic_data_service=topic_data_service,
                )
                for message_path_id in self.message_path_ids
            ]
            unique_topics = set(
                [message_path.topic_id for message_path in message_paths]
            )
            if len(unique_topics) > 1:
                raise RobotoInvalidRequestException(
                    "Unable to load event data for events associated with more than one topic"
                )

            # Request data using the message paths' containing topic,
            # which will better dedupe and make concurrent fetches for underlying data
            # across many message paths.
            topic_id = message_paths[0].topic_id
            topic = Topic.from_id(topic_id=topic_id, roboto_client=self.__roboto_client)
            yield from topic.get_data(
                message_paths_include=[
                    message_path.path for message_path in message_paths
                ],
                start_time=self.start_time,
                end_time=self.end_time,
                cache_dir=cache_dir,
            )
            return

        # Event associated with topic
        if len(self.topic_ids) > 0:
            unique_topics = set(self.topic_ids)
            if len(unique_topics) > 1:
                raise RobotoInvalidRequestException(
                    "Unable to load event data for events associated with more than one topic"
                )

            topic_id = self.topic_ids[0]
            topic = Topic.from_id(topic_id=topic_id, roboto_client=self.__roboto_client)
            yield from topic.get_data(
                message_paths_include=message_paths_include,
                message_paths_exclude=message_paths_exclude,
                start_time=self.start_time,
                end_time=self.end_time,
                cache_dir=cache_dir,
            )
            return

        # Event associated with a file
        if len(self.file_ids) != 0:
            if len(self.file_ids) > 1:
                raise RobotoInvalidRequestException(
                    "Unable to load event data for events associated with more than one file"
                )

            if topic_name is None:
                raise RobotoInvalidRequestException(
                    "Must provide 'topic_name' when attempting to load data for an event associated with a file"
                )

            file = File.from_id(
                file_id=self.file_ids[0], roboto_client=self.__roboto_client
            )
            topic = file.get_topic(topic_name)
            yield from topic.get_data(
                message_paths_include=message_paths_include,
                message_paths_exclude=message_paths_exclude,
                start_time=self.start_time,
                end_time=self.end_time,
                cache_dir=cache_dir,
            )
            return

        raise RobotoInvalidRequestException(
            "Can only load event data for events ultimately sourced from a single topic. "
            "That means it must be associated with either a single file, "
            "a single topic extracted from a file, or one or many specific message paths within a single topic."
        )

    def get_data_as_df(
        self,
        message_paths_include: typing.Optional[collections.abc.Sequence[str]] = None,
        message_paths_exclude: typing.Optional[collections.abc.Sequence[str]] = None,
        topic_name: typing.Optional[str] = None,
        topic_data_service: typing.Optional[TopicDataService] = None,
        cache_dir: typing.Union[str, pathlib.Path, None] = None,
    ) -> pandas.DataFrame:
        """
        Return the underlying topic data this event annotates as a pandas DataFrame.

        Requires installing this package using the ``roboto[analytics]`` extra.

        See :py:meth:`~roboto.domain.events.event.Event.get_data` for more information on the parameters.
        """
        pandas = import_optional_dependency("pandas", "analytics")

        df = pandas.json_normalize(
            data=list(
                self.get_data(
                    message_paths_include=message_paths_include,
                    message_paths_exclude=message_paths_exclude,
                    topic_name=topic_name,
                    topic_data_service=topic_data_service,
                    cache_dir=cache_dir,
                )
            )
        )

        if TopicDataService.LOG_TIME_ATTR_NAME not in df.columns:
            # Expected only in edge case:
            #   if this event's time bounds are outside the range of
            #   log times actually found in the underlying topic data.
            # In this case, the dataframe is expected to be entirely empty.
            return df

        return df.set_index(TopicDataService.LOG_TIME_ATTR_NAME)

    def put_metadata(self, metadata: dict[str, typing.Any]) -> "Event":
        return self.update(metadata_changeset=MetadataChangeset(put_fields=metadata))

    def put_tags(self, tags: list[str]) -> "Event":
        return self.update(metadata_changeset=MetadataChangeset(put_tags=tags))

    def refresh(self) -> "Event":
        self.__record = self.__roboto_client.get(
            f"v1/events/id/{self.event_id}"
        ).to_record(EventRecord)
        return self

    def remove_metadata(
        self,
        metadata: StrSequence,
    ) -> "Event":
        return self.update(metadata_changeset=MetadataChangeset(remove_fields=metadata))

    def remove_tags(
        self,
        tags: StrSequence,
    ) -> "Event":
        return self.update(metadata_changeset=MetadataChangeset(remove_tags=tags))

    def set_description(self, description: typing.Optional[str]) -> "Event":
        return self.update(description=description)

    def set_name(self, name: str) -> "Event":
        return self.update(name=name)

    def to_dict(self) -> dict[str, typing.Any]:
        return self.__record.model_dump(mode="json")

    def update(
        self,
        description: typing.Union[str, None, NotSetType] = NotSet,
        metadata_changeset: typing.Union[MetadataChangeset, NotSetType] = NotSet,
        name: typing.Union[str, NotSetType] = NotSet,
        start_time: typing.Union[int, NotSetType] = NotSet,
        end_time: typing.Union[int, NotSetType] = NotSet,
    ) -> "Event":
        request = remove_not_set(
            UpdateEventRequest(
                description=description,
                metadata_changeset=metadata_changeset,
                name=name,
                start_time=start_time,
                end_time=end_time,
            )
        )

        self.__record = self.__roboto_client.put(
            f"/v1/events/id/{self.event_id}", data=request
        ).to_record(EventRecord)

        return self



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

from __future__ import annotations

import collections.abc
import logging
import typing

import tqdm.auto

from ...compat import import_optional_dependency
from ...domain.topics import Topic
from ...logging import default_logger
from .match import Match, MatchContext

if typing.TYPE_CHECKING:
    import numpy  # pants: no-infer-dep
    import numpy.typing  # pants: no-infer-dep
    import pandas  # pants: no-infer-dep


logger = default_logger()

# Query signals must be at minimum 3 values long for results to be meaningful.
MIN_QUERY_LENGTH = 3


class MatchResult(typing.NamedTuple):
    start_idx: int
    end_idx: int
    distance: float


def _find_matches(
    query: numpy.typing.NDArray,
    target: numpy.typing.NDArray,
    *,
    max_distance: typing.Optional[float] = None,
    max_matches: typing.Optional[int] = None,
    normalize: bool = False,
) -> collections.abc.Sequence[MatchResult]:
    """
    For performing signal similarity, see :py:func:`~roboto.analytics.signal_similarity.find_similar_signals`.
    """
    stumpy = import_optional_dependency("stumpy", "analytics")

    if len(query) < MIN_QUERY_LENGTH:
        raise ValueError(
            f"Query sequence must be greater than {MIN_QUERY_LENGTH} for results to be meaningful. "
            f"Received sequence of length {len(query)}."
        )

    if len(query) > len(target):
        raise ValueError("Query sequence must be shorter than target")

    matches: list[MatchResult] = []
    for distance, start_idx in stumpy.match(
        query,
        target,
        max_distance=max_distance,
        max_matches=max_matches,
        normalize=normalize,
    ):
        end_idx = start_idx + len(query) - 1
        matches.append(
            MatchResult(
                start_idx=int(start_idx),
                end_idx=int(end_idx),
                distance=float(distance),
            )
        )

    return matches


def _find_matches_multidimensional(
    query: pandas.DataFrame,
    target: pandas.DataFrame,
    *,
    max_distance: typing.Optional[float] = None,
    max_matches: typing.Optional[int] = None,
    normalize: bool = False,
) -> collections.abc.Sequence[MatchResult]:
    """
    For performing signal similarity, see :py:func:`~roboto.analytics.signal_similarity.find_similar_signals`.
    """
    np = import_optional_dependency("numpy", "analytics")
    stumpy = import_optional_dependency("stumpy", "analytics")

    if len(query) < MIN_QUERY_LENGTH:
        raise ValueError(
            f"Query signal must be greater than {MIN_QUERY_LENGTH} for results to be meaningful. "
            f"Received DataFrame of size {query.shape}."
        )

    query_dims = set(query.columns.tolist())
    target_dims = set(target.columns.tolist())
    non_overlap = query_dims.difference(target_dims)
    if len(non_overlap):
        raise ValueError(
            "Cannot match query against target: they have non-overlapping dimensions. "
            f"Target signal is missing the following attributes: {non_overlap}"
        )

    # Accumulate summed distances for each subsequence (of length `query_signal`) within the target.
    # The distance for each subsequence starts at 0 and is incrementally updated.
    # Each dimension of the query signal (i.e., column in dataframe) is considered in turn:
    # for each dimension, compute the distance profile against the corresponding dimension in the target signal,
    # and then add that distance profile to the running total.
    # N.b.: for a target of len N and query of len M, there are a total of N - M + 1 subsequences
    #   (the first starts at index 0, the second at index 1, ..., the last starts at index N - M)
    summed_distance_profile: numpy.typing.NDArray[numpy.floating] = np.zeros(
        len(target) - len(query) + 1
    )
    for column in query_dims:
        query_sequence = query[column].to_numpy()
        target_sequence = target[column].to_numpy()
        distance_profile: numpy.typing.NDArray[numpy.floating] = stumpy.mass(
            query_sequence, target_sequence, normalize=normalize
        )
        summed_distance_profile += distance_profile

    matches: list[MatchResult] = []
    for distance, start_idx in stumpy.core._find_matches(
        summed_distance_profile,
        # https://github.com/TDAmeritrade/stumpy/blob/b7b355ce4a9450357ad207dd4f04fc8e8b4db100/stumpy/motifs.py#L533C17-L533C64
        excl_zone=int(np.ceil(len(query) / stumpy.core.config.STUMPY_EXCL_ZONE_DENOM)),
        max_distance=max_distance,
        max_matches=max_matches,
    ):
        end_idx = start_idx + len(query) - 1
        matches.append(
            MatchResult(
                start_idx=int(start_idx),
                end_idx=int(end_idx),
                distance=float(distance),
            )
        )

    return matches


def find_similar_signals(
    needle: pandas.DataFrame,
    haystack: collections.abc.Iterable[Topic],
    *,
    max_distance: typing.Optional[float] = None,
    max_matches_per_topic: typing.Optional[int] = None,
    normalize: bool = False,
) -> collections.abc.Sequence[Match]:
    """
    Find subsequences of topic data (from ``haystack``) that are similar to ``needle``.

    If ``needle`` is a dataframe with a single, non-index column,
    single-dimensional similarity search will be performed.
    If it instead has multiple non-index columns, multi-dimensional search will be performed.

    Even if there is no true similarity between the query signal and a topic's data,
    this will always return at least one :py:class:`~roboto.analytics.signal_similarity.Match`.
    Matches are expected to improve in quality as the topic data is more relevant to the query.
    Matches are returned sorted in ascending order by their distance, with the best matches (lowest distance) first.

    If ``max_distance`` is provided, only matches with a distance less than ``max_distance`` will be returned.
    Given distances computed against all comparison windows in the target, this defaults to the maximum of:
        1. the minimum distance
        2. the mean distance minus two standard deviations

    Use ``max_matches_per_topic`` to limit the number of match results contributed by a single topic.

    If ``normalize`` is True, values will be projected to the unit scale before matching.
    This is useful if you want to match windows of the target signal regardless of scale.
    For example, a query sequence of ``[1., 2., 3.]`` will perfectly match (distance == 0)
    the target ``[1000., 2000., 3000.]`` if ``normalize`` is True,
    but would have a distance of nearly 3800 if ``normalize`` is False.
    """
    matches: list[Match] = []
    _, cols = needle.shape

    targets = list(haystack)

    if cols == 1:
        # Single dimensional similarity search
        msg_path = needle.columns[0]
        query_sequence = needle[msg_path].to_numpy()
        for topic in tqdm.auto.tqdm(targets):
            match_context = MatchContext(
                dataset_id=topic.dataset_id,
                file_id=topic.file_id,
                message_paths=[msg_path],
                topic_name=topic.name,
                topic_id=topic.topic_id,
            )

            if logger.isEnabledFor(logging.DEBUG):
                tqdm.auto.tqdm.write(f"Loading data from {match_context!r}")

            topic_data = topic.get_data_as_df(message_paths_include=[msg_path])

            if logger.isEnabledFor(logging.DEBUG):
                tqdm.auto.tqdm.write(f"Searching for matches in {match_context!r}")

            target_signal = topic_data[msg_path].to_numpy()
            for match_result in _find_matches(
                query_sequence,
                target_signal,
                max_distance=max_distance,
                max_matches=max_matches_per_topic,
                normalize=normalize,
            ):
                matches.append(
                    Match(
                        context=match_context,
                        end_idx=match_result.end_idx,
                        end_time=topic_data.index[match_result.end_idx].item(),
                        distance=match_result.distance,
                        start_idx=match_result.start_idx,
                        start_time=topic_data.index[match_result.start_idx].item(),
                        subsequence=topic_data[
                            match_result.start_idx : match_result.end_idx + 1
                        ],
                    )
                )
    else:
        # Multi-dimensional match
        message_paths = needle.columns.tolist()

        for topic in tqdm.auto.tqdm(targets):
            match_context = MatchContext(
                dataset_id=topic.dataset_id,
                file_id=topic.file_id,
                message_paths=message_paths,
                topic_name=topic.name,
                topic_id=topic.topic_id,
            )

            if logger.isEnabledFor(logging.DEBUG):
                tqdm.auto.tqdm.write(f"Loading data from {match_context!r}")

            target_signal = topic.get_data_as_df(message_paths_include=message_paths)

            if logger.isEnabledFor(logging.DEBUG):
                tqdm.auto.tqdm.write(f"Searching for matches in {match_context!r}")

            for match_result in _find_matches_multidimensional(
                needle,
                target_signal,
                max_distance=max_distance,
                max_matches=max_matches_per_topic,
                normalize=normalize,
            ):
                matches.append(
                    Match(
                        context=match_context,
                        end_idx=match_result.end_idx,
                        end_time=target_signal.index[match_result.end_idx].item(),
                        distance=match_result.distance,
                        start_idx=match_result.start_idx,
                        start_time=target_signal.index[match_result.start_idx].item(),
                        subsequence=target_signal[
                            match_result.start_idx : match_result.end_idx + 1
                        ],
                    )
                )

    matches.sort(key=lambda match: match.distance)

    return matches



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import argparse
import json

from ...domain.orgs import Org
from ...domain.users import User
from ...exceptions import RobotoHttpExceptionParse
from ..command import (
    RobotoCommand,
    RobotoCommandSet,
)
from ..context import CLIContext


def show(args, context: CLIContext, parser: argparse.ArgumentParser):
    if not args.id:
        user = User.for_self(roboto_client=context.roboto_client)
    else:
        user = User.from_id(user_id=args.id, roboto_client=context.roboto_client)
    print(json.dumps(user.to_dict(), indent=2))


def show_setup_parser(parser):
    parser.add_argument(
        "--id",
        type=str,
        help=argparse.SUPPRESS,
    )


def delete(args, context: CLIContext, parser: argparse.ArgumentParser):
    if not args.ignore_prompt:
        print("Are you absolutely sure you want to delete your user? [y/n]:", end=" ")
        choice = input().lower()
        if choice not in ["y", "yes"]:
            return

    user = User.from_id(user_id=args.id, roboto_client=context.roboto_client)
    user.delete()
    print(f"Successfully deleted user '{args.id}'")


def delete_setup_parser(parser):
    parser.add_argument(
        "--id",
        type=str,
        help=argparse.SUPPRESS,
    )
    parser.add_argument(
        "--ignore-prompt",
        action="store_true",
        help="Ignore the prompt which asks you to confirm that you'd like to delete your user.",
    )


def orgs(args, context: CLIContext, parser: argparse.ArgumentParser):
    records = Org.for_self(roboto_client=context.roboto_client)
    for record in records:
        print(json.dumps(record.to_dict(), indent=2))


def whoami(args, context: CLIContext, parser: argparse.ArgumentParser):
    with RobotoHttpExceptionParse():
        contents = context.http_client.get(
            context.http_client.url("v1/users/whoami")
        ).to_dict(json_path=["data"])
        print(json.dumps(contents, indent=2))


delete_command = RobotoCommand(
    name="delete",
    logic=delete,
    setup_parser=delete_setup_parser,
    command_kwargs={"help": "Removes you from the roboto platform."},
)

orgs_command = RobotoCommand(
    name="orgs",
    logic=orgs,
    command_kwargs={"help": "Lists the roles that a user is a member of."},
)

show_command = RobotoCommand(
    name="show",
    logic=show,
    setup_parser=show_setup_parser,
    command_kwargs={"help": "Shows your user record."},
)

whoami_command = RobotoCommand(
    name="whoami",
    logic=whoami,
    command_kwargs={
        "help": "Returns the full identity context available to Roboto when you make a request."
    },
)

commands = [orgs_command, show_command, whoami_command, delete_command]

command_set = RobotoCommandSet(
    name="users",
    help="Get information about your account.",
    commands=commands,
)



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import pathlib
import typing

import pydantic

from roboto import sentinels
from roboto.domain import datasets, files
from roboto.env import RobotoEnv, RobotoEnvKey
from roboto.logging import default_logger
from roboto.updates import MetadataChangeset

logger = default_logger()


class FilesChangesetItem(pydantic.BaseModel):
    """
    A single row in the file metadata changeset .jsonl file.
    """

    relative_path: str
    """The path to the file relative to the ``${ROBOTO_OUTPUT_DIR}`` directory, which will correspond to the file's
    relative_path in a roboto dataset after upload."""

    update: files.UpdateFileRecordRequest
    """The file metadata update to apply to the file."""


class FilesChangesetFileManager:
    """
    This class is used to pre-write tags/metadata updates to files which haven't been uploaded yet, but will be at the
    conclusion of an action, by virtue of being in that action's output directory.

    It uses a "file changeset" file to accumulate these pending updates during an action's runtime, and then applies
    them automatically at the end of an action, after the action's output directory has been uploaded.

    The most common way to get access to this would be via :class:`roboto.action_runtime.ActionRuntime`.
    """

    __path: pathlib.Path

    def __init__(self):
        path_str = RobotoEnv.default().file_metadata_changeset_file
        if path_str is None:
            raise ValueError(
                "Couldn't locate expected file metadata changeset file from env var "
                + RobotoEnvKey.FileMetadataChangesetFile
            )

        path = pathlib.Path(path_str)
        if not path.exists():
            path.touch()

        self.__path = path

    def put_tags(self, relative_path: str, tags: list[str]):
        """
        Adds tags to a to-be-uploaded file which is expected to be written to
        ``${ROBOTO_OUTPUT_DIR}/relative_path`` by the end of the user portion of an action's runtime.

        This can be called multiple times throughout the runtime of an action, and will be accumulated accordingly.
        Order of calls matters.

        Examples:
            >>> from roboto import ActionRuntime
            >>> action_runtime = ActionRuntime.from_env()
            >>> file_changeset_manager = action_runtime.file_changeset_manager
            >>>
            >>> # This would reference a file at ${ROBOTO_OUTPUT_DIR}/images/front0_raw_000734.jpg
            >>> file_changeset_manager.put_tags("images/front0_raw_000734.jpg", ["cloudy", "rainy"]})
        """
        self.__append_update(
            relative_path,
            files.UpdateFileRecordRequest(
                metadata_changeset=MetadataChangeset(put_tags=tags)
            ),
        )

    def remove_tags(self, relative_path: str, tags: list[str]):
        """
        Removes tags from a to-be-uploaded file which is expected to be written to
        ``${ROBOTO_OUTPUT_DIR}/relative_path`` by the end of the user portion of an action's runtime. You'll generally
        only need this to remove tags which were added by a previous call to :meth:`put_tags`.

        This can be called multiple times throughout the runtime of an action, and will be accumulated accordingly.
        Order of calls matters.

        Examples:
            >>> from roboto import ActionRuntime
            >>> action_runtime = ActionRuntime.from_env()
            >>> file_changeset_manager = action_runtime.file_changeset_manager
            >>>
            >>> # This would reference a file at ${ROBOTO_OUTPUT_DIR}/images/front0_raw_000734.jpg
            >>> file_changeset_manager.put_tags("images/front0_raw_000734.jpg", ["cloudy", "rainy"]})
            >>>
            >>> # Actually this is just Seattle's aggressive mist, that's not really rainy...
            >>> file_changeset_manager.remove_tags("images/front0_raw_000734.jpg", ["rainy"]})
        """
        self.__append_update(
            relative_path,
            files.UpdateFileRecordRequest(
                metadata_changeset=MetadataChangeset(remove_tags=tags)
            ),
        )

    def put_fields(self, relative_path: str, metadata: dict[str, typing.Any]):
        """
        Adds metadata key/value pairs to a to-be-uploaded file which is expected to be written to
        ``${ROBOTO_OUTPUT_DIR}/relative_path`` by the end of the user portion of an action's runtime.

        This can be called multiple times throughout the runtime of an action, and will be accumulated accordingly.
        Order of calls matters.

        Examples:
            >>> from roboto import ActionRuntime
            >>> action_runtime = ActionRuntime.from_env()
            >>> file_changeset_manager = action_runtime.file_changeset_manager
            >>>
            >>> # This would reference a file at ${ROBOTO_OUTPUT_DIR}/images/front0_raw_000734.jpg
            >>> file_changeset_manager.put_fields("images/front0_raw_000734.jpg", {"cars": 2, "trucks": 3})
            >>>
            >>> # Actually there was a 3rd car I missed in the first pass, and a plane, let me fix that...
            >>> file_changeset_manager.put_fields("images/front0_raw_000734.jpg", {"cars": 3, "planes": 1})
        """
        self.__append_update(
            relative_path,
            files.UpdateFileRecordRequest(
                metadata_changeset=MetadataChangeset(put_fields=metadata)
            ),
        )

    def remove_fields(self, relative_path: str, keys: list[str]):
        """
        Removes metadata key/value pairs from a to-be-uploaded file which is expected to be written to
        ``${ROBOTO_OUTPUT_DIR}/relative_path`` by the end of the user portion of an action's runtime. You'll generally
        only need this to remove values which were added by a previous call to :meth:`put_fields`.

        This can be called multiple times throughout the runtime of an action, and will be accumulated accordingly.
        Order of calls matters.

        Examples:
            >>> from roboto import ActionRuntime
            >>> action_runtime = ActionRuntime.from_env()
            >>> file_changeset_manager = action_runtime.file_changeset_manager
            >>>
            >>> # This would reference a file at ${ROBOTO_OUTPUT_DIR}/images/front0_raw_000734.jpg
            >>> file_changeset_manager.put_fields("images/front0_raw_000734.jpg", {"cars": 2, "trucks": 3})
            >>>
            >>> # Whoops, actually I don't want to count those trucks...
            >>> file_changeset_manager.remove_fields("images/front0_raw_000734.jpg", ["trucks"])
        """
        self.__append_update(
            relative_path,
            files.UpdateFileRecordRequest(
                metadata_changeset=MetadataChangeset(remove_fields=keys)
            ),
        )

    def set_description(self, relative_path: str, description: typing.Optional[str]):
        """
        Sets the human-readable description of a to-be-uploaded file which is expected to be written to
        ``${ROBOTO_OUTPUT_DIR}/relative_path`` by the end of the user portion of an action's runtime.

        Examples:
            >>> from roboto import ActionRuntime
            >>> action_runtime = ActionRuntime.from_env()
            >>> file_changeset_manager = action_runtime.file_changeset_manager
            >>>
            >>> # This would reference a file at ${ROBOTO_OUTPUT_DIR}/images/front0_raw_000734.jpg
            >>> file_changeset_manager.set_description("images/front0_raw_000734.jpg", "This image was over-exposed")
        """
        self.__append_update(
            relative_path, files.UpdateFileRecordRequest(description=description)
        )

    def _apply_to_dataset(self, dataset: datasets.Dataset):
        """
        Applies all pending file metadata changes to the given dataset. This is what the Roboto action framework
        calls after the user portion of an action's runtime has completed. You should never need to call this directly.
        """
        if RobotoEnv.default().output_dir is None:
            raise ValueError(
                "Couldn't apply changes with no specified output dir via env var "
                + RobotoEnvKey.OutputDir
            )

        with open(self.__path, "r") as f:
            lines = f.readlines()

        try:
            items = [FilesChangesetItem.model_validate_json(line) for line in lines]
        except pydantic.ValidationError as ve:
            logger.error(
                "Got malformed files metadata changeset file, ignoring.", exc_info=ve
            )
            return

        updates_by_path: dict[str, files.UpdateFileRecordRequest] = {}

        for item in items:
            existing_update = updates_by_path.get(item.relative_path)

            if existing_update is None:
                updates_by_path[item.relative_path] = item.update
            else:
                new_update = existing_update

                if sentinels.is_set(item.update.description):
                    new_update.description = item.update.description
                if sentinels.is_set(
                    existing_update.metadata_changeset
                ) and sentinels.is_set(item.update.metadata_changeset):
                    new_update.metadata_changeset = (
                        existing_update.metadata_changeset.combine(
                            item.update.metadata_changeset
                        )
                    )

                updates_by_path[item.relative_path] = new_update

        relevant_files: list[files.File] = []

        for dataset_file in dataset.list_files():
            if dataset_file.relative_path in updates_by_path.keys():
                relevant_files.append(dataset_file)

        if len(relevant_files) == 0:
            logger.info("No file metadata update requests")
        else:
            logger.info(
                f"Attempting to update metadata for {len(relevant_files)} files"
            )

        for relevant_file in relevant_files:
            update_by_path = updates_by_path[relevant_file.relative_path]
            relevant_file.update(
                description=update_by_path.description,
                metadata_changeset=update_by_path.metadata_changeset,
            )

    def __append_update(
        self, relative_path: str, update: files.UpdateFileRecordRequest
    ):
        # Duplicates are OK, as we'll ignore them at upload time
        with open(self.__path, "a") as f:
            print(
                FilesChangesetItem(
                    relative_path=relative_path, update=update
                ).model_dump_json(),
                file=f,
            )



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

from .domain import (
    RobotoConditionException,
    RobotoConflictException,
    RobotoDeprecatedException,
    RobotoDomainException,
    RobotoExpiredException,
    RobotoHttpExceptionParse,
    RobotoIllegalArgumentException,
    RobotoInternalException,
    RobotoInvalidRequestException,
    RobotoInvalidStateTransitionException,
    RobotoLimitExceededException,
    RobotoNoOrgProvidedException,
    RobotoNotFoundException,
    RobotoNotImplementedException,
    RobotoNotReadyException,
    RobotoServiceException,
    RobotoServiceUnavailableException,
    RobotoUnauthorizedException,
    RobotoUnknownOperationException,
)
from .http import (
    ClientError,
    HttpError,
    ServerError,
)

__all__ = [
    "ClientError",
    "HttpError",
    "ServerError",
    "RobotoConditionException",
    "RobotoConflictException",
    "RobotoDeprecatedException",
    "RobotoDomainException",
    "RobotoExpiredException",
    "RobotoHttpExceptionParse",
    "RobotoIllegalArgumentException",
    "RobotoInternalException",
    "RobotoInvalidRequestException",
    "RobotoInvalidStateTransitionException",
    "RobotoLimitExceededException",
    "RobotoNotFoundException",
    "RobotoNotImplementedException",
    "RobotoNotReadyException",
    "RobotoNoOrgProvidedException",
    "RobotoServiceException",
    "RobotoServiceUnavailableException",
    "RobotoUnauthorizedException",
    "RobotoUnknownOperationException",
]



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

from typing import Optional

from .constants import (
    ORG_OVERRIDE_HEADER,
    RESOURCE_OWNER_OVERRIDE_HEADER,
    USER_OVERRIDE_HEADER,
)

CONTENT_TYPE_JSON_HEADER = {"Content-Type": "application/json"}


def roboto_headers(
    org_id: Optional[str] = None,
    user_id: Optional[str] = None,
    resource_owner_id: Optional[str] = None,
    additional_headers: Optional[dict[str, str]] = None,
):
    headers = {}

    if org_id is not None:
        headers[ORG_OVERRIDE_HEADER] = org_id

    if user_id is not None:
        headers[USER_OVERRIDE_HEADER] = user_id

    if resource_owner_id is not None:
        headers[RESOURCE_OWNER_OVERRIDE_HEADER] = resource_owner_id

    if additional_headers is not None:
        headers.update(additional_headers)

    return headers



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

from .action import Action
from .action_operations import (
    CreateActionRequest,
    SetActionAccessibilityRequest,
    UpdateActionRequest,
)
from .action_record import (
    Accessibility,
    ActionParameter,
    ActionParameterChangeset,
    ActionRecord,
    ActionReference,
    ComputeRequirements,
    ContainerParameters,
    ExecutorContainer,
)
from .invocation import Invocation
from .invocation_operations import (
    CreateInvocationRequest,
    SetContainerInfoRequest,
    SetLogsLocationRequest,
    UpdateInvocationStatus,
)
from .invocation_record import (
    ActionProvenance,
    ExecutableProvenance,
    InvocationDataSource,
    InvocationDataSourceType,
    InvocationProvenance,
    InvocationRecord,
    InvocationSource,
    InvocationStatus,
    InvocationStatusRecord,
    LogRecord,
    LogsLocation,
    SourceProvenance,
)
from .trigger import Trigger
from .trigger_operations import (
    CreateTriggerRequest,
    EvaluateTriggersRequest,
    QueryTriggersRequest,
    TriggerEvaluationsSummaryResponse,
    UpdateTriggerRequest,
)
from .trigger_record import (
    TriggerEvaluationCause,
    TriggerEvaluationOutcome,
    TriggerEvaluationOutcomeReason,
    TriggerEvaluationRecord,
    TriggerEvaluationStatus,
    TriggerForEachPrimitive,
    TriggerRecord,
)

__all__ = (
    "Accessibility",
    "Action",
    "ActionParameter",
    "ActionParameterChangeset",
    "ActionProvenance",
    "ActionRecord",
    "ActionReference",
    "ComputeRequirements",
    "ContainerParameters",
    "CreateActionRequest",
    "CreateInvocationRequest",
    "CreateTriggerRequest",
    "EvaluateTriggersRequest",
    "ExecutableProvenance",
    "ExecutorContainer",
    "Invocation",
    "InvocationDataSource",
    "InvocationDataSourceType",
    "InvocationProvenance",
    "InvocationRecord",
    "InvocationSource",
    "InvocationStatus",
    "InvocationStatusRecord",
    "LogsLocation",
    "LogRecord",
    "QueryTriggersRequest",
    "SetActionAccessibilityRequest",
    "SetContainerInfoRequest",
    "SetLogsLocationRequest",
    "SourceProvenance",
    "Trigger",
    "TriggerEvaluationCause",
    "TriggerEvaluationOutcome",
    "TriggerEvaluationOutcomeReason",
    "TriggerEvaluationRecord",
    "TriggerEvaluationStatus",
    "TriggerEvaluationsSummaryResponse",
    "TriggerForEachPrimitive",
    "TriggerRecord",
    "UpdateActionRequest",
    "UpdateInvocationStatus",
    "UpdateTriggerRequest",
)



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

from ..command import RobotoCommandSet
from .create import create_command
from .delete_dataset import delete_dataset_command
from .delete_files import delete_files_command
from .download_files import download_files_command
from .list_files import list_files_command
from .search import search_command
from .show import show_command
from .update import update_command
from .upload_files import upload_files_command

commands = [
    create_command,
    delete_dataset_command,
    delete_files_command,
    download_files_command,
    list_files_command,
    show_command,
    search_command,
    update_command,
    upload_files_command,
]

command_set = RobotoCommandSet(
    name="datasets",
    help="Manage data from a single event, such a robot run or drone flight. Includes file upload and download.",
    commands=commands,
)



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import argparse
import shlex
import subprocess
import sys

from ...auth import Permissions
from ...image_registry import ImageRegistry
from ..command import RobotoCommand
from ..common_args import add_org_arg
from ..context import CLIContext


def pull(
    args: argparse.Namespace, context: CLIContext, parser: argparse.ArgumentParser
) -> None:
    image_registry = ImageRegistry(context.roboto_client)
    parts = args.remote_image.split(":")
    if len(parts) == 1:
        repository_uri = parts[0]
    elif len(parts) == 2:
        repository_uri, _ = parts
    else:
        raise ValueError("Invalid image format. Expected '<repository>:<tag>'.")

    credentials = image_registry.get_temporary_credentials(
        repository_uri, Permissions.ReadOnly, org_id=args.org
    )
    login_cmd = f"docker login --username {credentials.username} --password-stdin {credentials.registry_url}"
    try:
        subprocess.run(
            shlex.split(login_cmd),
            capture_output=True,
            check=True,
            input=credentials.password,
            text=True,
        )
    except subprocess.CalledProcessError as exc:
        print(
            "Failed to set Docker credentials for Roboto's image registry.",
            file=sys.stderr,
        )
        if exc.stdout:
            print(exc.stdout, file=sys.stderr)
        if exc.stderr:
            print(exc.stderr, file=sys.stderr)
        return

    pull_cmd = f"docker pull {args.remote_image}"
    with subprocess.Popen(
        shlex.split(pull_cmd),
        text=True,
    ) as pull_proc:
        try:
            pull_proc.wait()
        except KeyboardInterrupt:
            pull_proc.kill()
            return


def pull_parser(parser: argparse.ArgumentParser) -> None:
    add_org_arg(parser)

    parser.add_argument(
        "remote_image",
        action="store",
        help="Specify the remote image to pull, in the format ``<repository>:<tag>``.",
    )


pull_command = RobotoCommand(
    name="pull",
    logic=pull,
    setup_parser=pull_parser,
    command_kwargs={
        "help": (
            "Pull a container image hosted in Roboto's image registry. "
            "Requires Docker CLI."
        )
    },
)



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import pydantic


class AuthZTupleRecord(pydantic.BaseModel):
    """
    Fully qualified record of (user has relation to obj)
    """

    user: str
    relation: str
    obj: str


class EditAccessRequest(pydantic.BaseModel):
    """
    Request payload to add or remove fine-grained access to a Roboto resource
    """

    add: list[AuthZTupleRecord] = pydantic.Field(default_factory=list)
    remove: list[AuthZTupleRecord] = pydantic.Field(default_factory=list)


class GetAccessResponse(pydantic.BaseModel):
    """
    Response payload for a request to describe fine-grained access to a Roboto resource
    """

    relations: list[AuthZTupleRecord]
    group_permissions: dict[str, list[str]] = pydantic.Field(default_factory=dict)



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import datetime

import pydantic


class DeviceRecord(pydantic.BaseModel):
    """
    A wire-transmissible representation of a device.
    """

    created: datetime.datetime = pydantic.Field(
        description="Date/time when this device was registered."
    )
    created_by: str = pydantic.Field(description="The user who registered this device.")
    device_id: str = pydantic.Field(
        description="A user-provided identifier for a device, which is unique within that device's org."
    )
    modified: datetime.datetime = pydantic.Field(
        description="Date/time when this device record was last modified."
    )
    modified_by: str = pydantic.Field(
        description="The user who last modified this device record."
    )
    org_id: str = pydantic.Field(description="The org to which this device belongs.")



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import pathlib
import typing

import pydantic
from pydantic import ConfigDict

from ...domain import actions


class DockerImageConfig(pydantic.BaseModel):
    dockerfile: pathlib.Path = pathlib.Path.cwd() / "Dockerfile"
    context: pathlib.Path = pathlib.Path.cwd()
    build_args: dict[str, str] = pydantic.Field(default_factory=dict)


class ActionConfig(pydantic.BaseModel):
    """Model for file-based Action config, to be used when creating or updating Actions."""

    # Required
    name: str

    # Optional
    compute_requirements: typing.Optional[actions.ComputeRequirements] = None
    container_parameters: typing.Optional[actions.ContainerParameters] = None
    description: typing.Optional[str] = None
    inherits: typing.Optional[actions.ActionReference] = None
    metadata: dict[str, typing.Any] = pydantic.Field(default_factory=dict)
    parameters: list[actions.ActionParameter] = pydantic.Field(default_factory=list)
    tags: list[str] = pydantic.Field(default_factory=list)
    short_description: typing.Optional[str] = None
    timeout: typing.Optional[int] = None

    # Mutually exclusive
    docker_config: typing.Optional[DockerImageConfig] = None
    """Configuration with which the CLI can build a Docker image for the Action."""

    image_uri: typing.Optional[str] = None
    """URI to a non-local Docker image. Must already be pushed a registry accessible by the Roboto Platform."""

    @pydantic.model_validator(mode="before")
    def enforce_invariants(cls, values: dict) -> dict:
        if values.get("docker_config") and values.get("image_uri"):
            raise ValueError(
                "'docker_config' and 'image_uri' are mutually exclusive configuration options. "
                "Use 'docker_config' to build a Docker image from a local Dockerfile, "
                "or 'image_uri' to reference a Docker image already pushed to a registry "
                "accessible by the Roboto Platform."
            )
        return values

    model_config = ConfigDict(extra="forbid")



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

from ..command import RobotoCommandSet
from .create import create_command
from .delete import delete_command
from .init import init_command
from .invoke import invoke_command
from .list_invocations import (
    list_invocations_command,
)
from .search import search_command
from .show import show_command
from .update import update_command

commands = [
    create_command,
    delete_command,
    update_command,
    search_command,
    show_command,
    invoke_command,
    list_invocations_command,
    init_command,
]

command_set = RobotoCommandSet(
    name="actions",
    help=(
        "Create, edit and invoke reusable actions that run containerized code on your datasets."
    ),
    commands=commands,
)



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import argparse
import os
import pathlib

from cookiecutter.main import cookiecutter

from ..command import RobotoCommand
from ..context import CLIContext

COOKIECUTTER_REPO = "https://github.com/roboto-ai/cookiecutter-roboto-actions.git"


def init(
    args: argparse.Namespace, context: CLIContext, parser: argparse.ArgumentParser
) -> None:
    path: pathlib.Path = args.path
    path = path.resolve()
    print(f"Creating Roboto action package under: {path}")

    if not path.is_dir():
        parser.error(f"{path} does not exist, or is not a directory.")

    cookiecutter(COOKIECUTTER_REPO, output_dir=path)


def init_parser(parser: argparse.ArgumentParser):
    parser.add_argument(
        "path",
        nargs="?",
        type=pathlib.Path,
        help="Existing directory under which an action package will be created.",
        default=os.getcwd(),
    )


init_command = RobotoCommand(
    name="init",
    logic=init,
    setup_parser=init_parser,
    command_kwargs={"help": "Initialize a new action package."},
)



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import argparse
import collections.abc
import logging
import pathlib
import sys
import time

import filelock
import pydantic

from ..config import DEFAULT_ROBOTO_DIR
from ..domain import datasets
from ..http import (
    RobotoClient,
    RobotoRequester,
    RobotoTool,
)
from ..logging import default_logger
from .agent import UploadAgent
from .files import UploadAgentConfig

logger = default_logger()
logging.basicConfig(
    format="%(asctime)s [%(levelname)s]: %(message)s", datefmt="%Y-%m-%dT%H:%M:%S%z"
)

agent_config_file = DEFAULT_ROBOTO_DIR / "upload_agent.json"
agent_config_lockfile = DEFAULT_ROBOTO_DIR / "tmp" / "upload_agent.lock"


def configure():
    if not DEFAULT_ROBOTO_DIR.is_dir():
        print(
            f"Roboto config directory {DEFAULT_ROBOTO_DIR} has not been created yet, creating it now\n",
        )
        DEFAULT_ROBOTO_DIR.mkdir(parents=True)

    if agent_config_file.is_file():
        print(
            f"Roboto config file {agent_config_file} already exists, this will command overwrite it.",
        )
        choice = input("Do you want to continue? [y/n]: ").lower()
        if choice not in ["y", "yes"]:
            return
        print("")

    print(
        "Enter an upload config filename for the agent to search for, or press enter to use the "
        + "default (.roboto_upload.json).\n"
        + "Any file with this name signals to the upload agent that a new dataset should be created and files in the "
        + "same directory or subdirectories as the upload config file should be uploaded to the newly created dataset."
    )
    upload_config_filename = input("Upload config filename [.roboto_upload.json]: ")
    upload_config_filename = upload_config_filename.strip() or ".roboto_upload.json"

    print("")
    choice = input(
        "Do you want the roboto-agent to locally delete files after they've been uploaded? [y/n]: "
    ).lower()
    delete_uploaded_files = choice in ["y", "yes"]

    search_paths: list[pathlib.Path] = []
    print(
        "\nEnter one search path at a time (absolute or relative).\n"
        "These paths will be scanned recursively for upload config files each time the upload agent runs.\n"
        "Press enter on an empty line to finish entering search paths.\n"
    )

    while True:
        user_input = input("Search path: ")
        if user_input.strip() == "":
            if len(search_paths) == 0:
                print(
                    "No search paths were provided, at least one is required for the upload agent to work.\n",
                )
                continue
            print("Done adding search paths.\n")
            break

        search_path = pathlib.Path(user_input).resolve()
        if not search_path.is_dir():
            print(
                f"Path {search_path} does not exist or is not a directory.\n",
            )
            continue

        search_paths.append(search_path)
        print(f"Added search path {search_path}\n")

    new_config_file = UploadAgentConfig(
        upload_config_filename=upload_config_filename,
        search_paths=search_paths,
        delete_uploaded_files=delete_uploaded_files,
    )

    agent_config_file.write_text(new_config_file.model_dump_json(indent=2))
    print(f"Wrote config file to {agent_config_file}")


def configure_subcommand(args: argparse.Namespace) -> None:
    configure()


def run(auto_create_upload_configs: bool, merge_uploads: bool) -> None:
    if not agent_config_file.is_file():
        logger.error(
            f"No upload agent config file found at {agent_config_file}. Please run "
            + "'roboto-agent configure' to generate it."
        )
        return

    try:
        agent_config = UploadAgentConfig.model_validate_json(
            agent_config_file.read_text()
        )
    except pydantic.ValidationError:
        logger.error(
            f"Upload agent config file {agent_config_file} could not be parsed, which means it's incorrectly "
            + "formatted. Please run 'roboto-agent configure' to generate a new one."
        )
        return

    roboto_client = RobotoClient.from_env()
    roboto_client.http_client.set_requester(
        RobotoRequester.for_tool(RobotoTool.UploadAgent)
    )

    upload_agent = UploadAgent(agent_config)

    uploaded_datasets: collections.abc.Sequence[datasets.Dataset]

    try:
        with filelock.FileLock(agent_config_lockfile, timeout=1):
            if auto_create_upload_configs:
                upload_agent.create_upload_configs()

            uploaded_datasets = upload_agent.process_uploads(
                merge_uploads=merge_uploads
            )
    except filelock.Timeout:
        logger.info(
            "Roboto upload agent appears to already be running, nothing to do. If you don't think this is correct, "
            + "the agent filelock %s may have been orphaned, which you can fix by deleting it.",
            agent_config_lockfile,
        )
        return

    logger.info("Uploaded %d datasets", len(uploaded_datasets))


def run_forever(
    scan_period_seconds: int, auto_create_upload_configs: bool, merge_uploads: bool
) -> None:
    print(
        "Starting roboto-agent in run forever mode, press Ctrl+C to stop.",
        file=sys.stdout,
    )

    try:
        while True:
            logger.info("Running upload agent")
            run(
                auto_create_upload_configs=auto_create_upload_configs,
                merge_uploads=merge_uploads,
            )
            logger.info(
                f"Run completed, sleeping for {scan_period_seconds} seconds before next attempt."
            )
            time.sleep(scan_period_seconds)
    except KeyboardInterrupt:
        pass


def run_subcommand(args: argparse.Namespace) -> None:
    if args.forever:
        run_forever(
            scan_period_seconds=30,
            auto_create_upload_configs=args.auto_create_upload_configs,
            merge_uploads=args.merge_uploads,
        )
    else:
        run(
            auto_create_upload_configs=args.auto_create_upload_configs,
            merge_uploads=args.merge_uploads,
        )


def main():
    parser = argparse.ArgumentParser(
        prog="roboto-agent",
        description="Upload agent for managing automatic dataset uploads to Roboto based on configuration.",
    )

    verbosity_group = parser.add_mutually_exclusive_group()
    verbosity_group.add_argument(
        "--verbose",
        "-v",
        help=(
            "Set increasing levels of verbosity. "
            "-v prints WARNING logs, -vv prints INFO logs, -vvv prints DEBUG logs."
        ),
        action="count",
        default=0,
    )
    verbosity_group.add_argument(
        "-q", "--quiet", action="store_true", help="Suppress output except for errors"
    )

    subparsers = parser.add_subparsers(dest="command", help="sub-command help")

    run_parser = subparsers.add_parser("run")
    run_parser.add_argument(
        "-f",
        "--forever",
        help="Attempts to call run every 30 seconds forever, "
        + "and sleeps between runs.",
        action="store_true",
    )
    run_parser.add_argument(
        "-m",
        "--merge-uploads",
        action="store_true",
        help=(
            "If set, all uploads will be merged into a single dataset. If combined with "
            "--auto-create-upload-configs, this will allow you to set many disparate output locations as your "
            "search paths, and still unite everything under a single dataset. "
            "Any tags/metadata/description set in .roboto_upload.json files will be applied sequentially as updates "
            "to the created dataset. If there are collisions, like multiple different descriptions or multiple "
            "metadata values for the same key, the last one encountered will be used, and the traversal order will "
            "be non-deterministic."
        ),
    )
    run_parser.add_argument(
        "-a",
        "--auto-create-upload-configs",
        action="store_true",
        help="If set, the agent will attempt to create upload configs for any directories one level under your "
        + "configured search paths which don't already have a .roboto_upload.json, "
        + ".roboto_upload_in_progress.json, or .roboto_upload_complete.json file. "
        + "For example, if your search_paths are set to ['/home/username/output'], a directory like "
        + "/home/username/output/2024_01_01T00_00_00 will be automatically prepared for upload. "
        + "By default the contents of the generated .roboto_upload.json file will be {}, but you can override "
        + "this by writing a ~/.roboto/default_roboto_upload.json file, whose contents will always be used instead. "
        + "Environment variables referenced in this file will be resolved appropriately, so you can add something "
        + 'like {"dataset":{"metadata":{"user_from_env":"$USER"}}}.',
    )
    run_parser.set_defaults(func=run_subcommand)

    configure_parser = subparsers.add_parser("configure")
    configure_parser.set_defaults(func=configure_subcommand)

    args = parser.parse_args()

    if args.quiet:
        log_level = logging.ERROR
    elif args.verbose >= 3:
        log_level = logging.DEBUG
    elif args.verbose == 2:
        log_level = logging.INFO
    elif args.verbose == 1:
        log_level = logging.WARNING

    # Default to INFO if no verbosity or quiet flags are provided
    else:
        log_level = logging.INFO

    logging.getLogger().setLevel(log_level)

    if args.command is None:
        parser.print_help()
        return

    args.func(args)


if __name__ == "__main__":
    main()



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import argparse
import json

from ...domain import actions
from ..command import RobotoCommand
from ..common_args import (
    add_action_reference_arg,
    add_org_arg,
)
from ..context import CLIContext


def show(
    args: argparse.Namespace, context: CLIContext, parser: argparse.ArgumentParser
) -> None:
    owner_org_id = args.action.owner if args.action.owner else args.org
    action = actions.Action.from_name(
        name=args.action.name,
        digest=args.action.digest,
        owner_org_id=owner_org_id,
        roboto_client=context.roboto_client,
    )
    print(json.dumps(action.to_dict(), indent=2))


def show_parser(parser: argparse.ArgumentParser):
    add_action_reference_arg(parser)
    add_org_arg(parser=parser)


show_command = RobotoCommand(
    name="show",
    logic=show,
    setup_parser=show_parser,
    command_kwargs={"help": "Show details for an action."},
)



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

from .org import Org
from .org_invite import OrgInvite
from .org_operations import (
    BindEmailDomainRequest,
    CreateOrgRequest,
    InviteUserRequest,
    ModifyRoleForUserRequest,
    RemoveUserFromOrgRequest,
    UpdateOrgRequest,
    UpdateOrgUserRequest,
)
from .org_records import (
    OrgInviteRecord,
    OrgRecord,
    OrgRoleName,
    OrgStatus,
    OrgTier,
    OrgUserRecord,
)

__all__ = [
    "BindEmailDomainRequest",
    "CreateOrgRequest",
    "InviteUserRequest",
    "ModifyRoleForUserRequest",
    "Org",
    "OrgInvite",
    "OrgInviteRecord",
    "OrgRecord",
    "OrgRoleName",
    "OrgStatus",
    "OrgTier",
    "OrgUserRecord",
    "RemoveUserFromOrgRequest",
    "UpdateOrgRequest",
    "UpdateOrgUserRequest",
]



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import collections.abc
import logging
import typing
import urllib.parse

from ..config import RobotoConfig
from ..exceptions import RobotoHttpExceptionParse
from ..logging import LOGGER_NAME
from .headers import roboto_headers
from .http_client import HttpClient
from .request import HttpRequestDecorator
from .request_decorators import (
    BearerTokenDecorator,
)
from .requester import RobotoRequester, RobotoTool
from .response import HttpResponse
from .retry import RetryWaitFn

logger = logging.getLogger(LOGGER_NAME)


ApiRelativePath = typing.Union[str, collections.abc.Sequence[str]]


class RobotoClient:
    """
    A client for making HTTP requests against Roboto service
    """

    __from_env_instance: typing.ClassVar[typing.Optional["RobotoClient"]] = None
    __endpoint: str
    __http_client: HttpClient

    @classmethod
    def from_config(cls, config: RobotoConfig) -> "RobotoClient":
        auth_decorator = BearerTokenDecorator(token=config.api_key)
        return RobotoClient(endpoint=config.endpoint, auth_decorator=auth_decorator)

    @classmethod
    def from_env(cls) -> "RobotoClient":
        if cls.__from_env_instance is None:
            cls.__from_env_instance = RobotoClient.from_config(RobotoConfig.from_env())

            # We know we're in the SDK right now, the CLI or Upload Agent will explicitly re-call set_requester
            # after this is returned if they're the more correct value for tool.
            cls.__from_env_instance.http_client.set_requester(
                RobotoRequester.for_tool(RobotoTool.Sdk)
            )

        return cls.__from_env_instance

    @classmethod
    def defaulted(
        cls, client: typing.Optional["RobotoClient"] = None
    ) -> "RobotoClient":
        return client or RobotoClient.from_env()

    def __init__(
        self, endpoint: str, auth_decorator: typing.Optional[HttpRequestDecorator]
    ):
        self.__endpoint = endpoint
        self.__http_client = HttpClient(
            default_endpoint=endpoint, default_auth=auth_decorator
        )

    @property
    def http_client(self) -> HttpClient:
        return self.__http_client

    @property
    def endpoint(self) -> str:
        return self.__endpoint

    @property
    def frontend_endpoint(self) -> str:
        return self.__endpoint.replace("api", "app")

    def delete(
        self,
        path: ApiRelativePath,
        caller_org_id: typing.Optional[str] = None,
        data: typing.Any = None,
        headers: typing.Optional[dict[str, str]] = None,
        idempotent: bool = True,
        owner_org_id: typing.Optional[str] = None,
        query: typing.Optional[dict[str, typing.Any]] = None,
        retry_wait_fn: typing.Optional[RetryWaitFn] = None,
    ) -> HttpResponse:
        with RobotoHttpExceptionParse():
            return self.__http_client.delete(
                url=self.__build_url(path, query),
                data=data,
                headers=roboto_headers(
                    org_id=caller_org_id,
                    resource_owner_id=owner_org_id,
                    additional_headers=headers,
                ),
                retry_wait=retry_wait_fn,
                idempotent=idempotent,
            )

    def get(
        self,
        path: ApiRelativePath,
        caller_org_id: typing.Optional[str] = None,
        headers: typing.Optional[dict[str, str]] = None,
        idempotent: bool = True,
        owner_org_id: typing.Optional[str] = None,
        query: typing.Optional[dict[str, typing.Any]] = None,
        retry_wait_fn: typing.Optional[RetryWaitFn] = None,
    ) -> HttpResponse:
        with RobotoHttpExceptionParse():
            return self.__http_client.get(
                url=self.__build_url(path, query),
                headers=roboto_headers(
                    org_id=caller_org_id,
                    resource_owner_id=owner_org_id,
                    additional_headers=headers,
                ),
                retry_wait=retry_wait_fn,
                idempotent=idempotent,
            )

    def patch(
        self,
        path: ApiRelativePath,
        caller_org_id: typing.Optional[str] = None,
        data: typing.Any = None,
        headers: typing.Optional[dict[str, str]] = None,
        idempotent: bool = True,
        owner_org_id: typing.Optional[str] = None,
        query: typing.Optional[dict[str, typing.Any]] = None,
        retry_wait_fn: typing.Optional[RetryWaitFn] = None,
    ) -> HttpResponse:
        with RobotoHttpExceptionParse():
            return self.__http_client.patch(
                url=self.__build_url(path, query),
                data=data,
                headers=roboto_headers(
                    org_id=caller_org_id,
                    resource_owner_id=owner_org_id,
                    additional_headers=headers,
                ),
                retry_wait=retry_wait_fn,
                idempotent=idempotent,
            )

    def post(
        self,
        path: ApiRelativePath,
        caller_org_id: typing.Optional[str] = None,
        data: typing.Any = None,
        headers: typing.Optional[dict[str, str]] = None,
        idempotent: bool = True,
        owner_org_id: typing.Optional[str] = None,
        query: typing.Optional[dict[str, typing.Any]] = None,
        retry_wait_fn: typing.Optional[RetryWaitFn] = None,
    ) -> HttpResponse:
        with RobotoHttpExceptionParse():
            return self.__http_client.post(
                url=self.__build_url(path, query),
                data=data,
                headers=roboto_headers(
                    org_id=caller_org_id,
                    resource_owner_id=owner_org_id,
                    additional_headers=headers,
                ),
                retry_wait=retry_wait_fn,
                idempotent=idempotent,
            )

    def put(
        self,
        path: ApiRelativePath,
        caller_org_id: typing.Optional[str] = None,
        data: typing.Any = None,
        headers: typing.Optional[dict[str, str]] = None,
        idempotent: bool = True,
        owner_org_id: typing.Optional[str] = None,
        query: typing.Optional[dict[str, typing.Any]] = None,
        retry_wait_fn: typing.Optional[RetryWaitFn] = None,
    ) -> HttpResponse:
        with RobotoHttpExceptionParse():
            return self.__http_client.put(
                url=self.__build_url(path, query),
                data=data,
                headers=roboto_headers(
                    org_id=caller_org_id,
                    resource_owner_id=owner_org_id,
                    additional_headers=headers,
                ),
                retry_wait=retry_wait_fn,
                idempotent=idempotent,
            )

    def __build_url(
        self,
        path: ApiRelativePath,
        query: typing.Optional[dict[str, typing.Any]] = None,
    ) -> str:
        if isinstance(path, str):
            normalized_path = path
        else:
            normalized_path = "/".join(path)

        normalized_path = normalized_path.lstrip("/")

        if not query:
            return f"{self.__endpoint}/{normalized_path}"
        else:
            return (
                f"{self.__endpoint}/{normalized_path}?{urllib.parse.urlencode(query)}"
            )



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

from .entry import construct_parser, entry

__version__ = "0.0.1"

__all__ = ["__version__", "entry", "construct_parser"]



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import contextlib
import typing

import pydantic

from .terminal import print_error_and_exit


class pydantic_validation_handler(contextlib.AbstractContextManager):
    """
    Context manager to catch Pydantic validation errors and print the first error neatly to stderr.
    """

    __parseable_name: str = "Roboto model"

    def __init__(self, parseable_name: typing.Optional[str] = None):
        if parseable_name is not None:
            self.__parseable_name = parseable_name

    def __enter__(self):
        pass

    def __exit__(self, exctype, excinst, exctb):
        if exctype is None:
            return

        if issubclass(exctype, pydantic.ValidationError):
            first_error, *_rest = excinst.errors()
            msg = first_error.get("msg")
            loc = first_error.get("loc", tuple())
            if loc:
                path_to_error = ".".join([str(key_path) for key_path in loc])
                msg = f"'{path_to_error}' {msg}"

            error_msg = [f"Error parsing {self.__parseable_name}"]
            if msg:
                error_msg.append(msg)

            print_error_and_exit(error_msg)

        return False



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import collections.abc
import datetime
import typing

import pydantic
from pydantic import ConfigDict

from roboto.sentinels import NotSet, NotSetType

from . import (
    ComputeRequirements,
    ContainerParameters,
)
from ...pydantic import (
    validate_nonzero_gitpath_specs,
)
from ...query import ConditionType
from .trigger_record import (
    TriggerForEachPrimitive,
)


class CreateTriggerRequest(pydantic.BaseModel):
    """
    Request payload to create a new trigger
    """

    action_digest: typing.Optional[str] = None
    action_name: str
    action_owner_id: typing.Optional[str] = None
    additional_inputs: typing.Optional[list[str]] = None
    compute_requirement_overrides: typing.Optional[ComputeRequirements] = None
    condition: typing.Optional[ConditionType] = None
    container_parameter_overrides: typing.Optional[ContainerParameters] = None
    enabled: bool = True
    for_each: TriggerForEachPrimitive
    name: str = pydantic.Field(pattern=r"[\w\-]+", max_length=256)
    parameter_values: typing.Optional[dict[str, typing.Any]] = None
    required_inputs: list[str]
    service_user_id: typing.Optional[str] = None
    timeout: typing.Optional[int] = None

    @pydantic.field_validator("required_inputs")
    def validate_required_inputs(cls, value: list[str]) -> list[str]:
        return validate_nonzero_gitpath_specs(value)

    @pydantic.field_validator("additional_inputs")
    def validate_additional_inputs(
        cls, value: typing.Optional[list[str]]
    ) -> typing.Optional[list[str]]:
        return None if value is None else validate_nonzero_gitpath_specs(value)


class EvaluateTriggersRequest(pydantic.BaseModel):
    """
    Request payload to evaluate triggers
    """

    trigger_evaluation_ids: collections.abc.Iterable[int]


class QueryTriggersRequest(pydantic.BaseModel):
    """
    Request payload to query triggers
    """

    filters: dict[str, typing.Any] = pydantic.Field(default_factory=dict)
    model_config = ConfigDict(extra="forbid")


class TriggerEvaluationsSummaryResponse(pydantic.BaseModel):
    """
    Response to trigger evaluations summary
    """

    count_pending: int
    last_evaluation_start: typing.Optional[datetime.datetime]


class UpdateTriggerRequest(pydantic.BaseModel):
    """
    Request payload to update a trigger
    """

    action_name: typing.Union[str, NotSetType] = NotSet
    action_owner_id: typing.Union[str, NotSetType] = NotSet
    action_digest: typing.Optional[typing.Union[str, NotSetType]] = NotSet
    additional_inputs: typing.Optional[typing.Union[list[str], NotSetType]] = NotSet
    compute_requirement_overrides: typing.Optional[
        typing.Union[ComputeRequirements, NotSetType]
    ] = NotSet
    container_parameter_overrides: typing.Optional[
        typing.Union[ContainerParameters, NotSetType]
    ] = NotSet
    condition: typing.Optional[typing.Union[ConditionType, NotSetType]] = NotSet
    enabled: typing.Union[bool, NotSetType] = NotSet
    for_each: typing.Union[TriggerForEachPrimitive, NotSetType] = NotSet
    parameter_values: typing.Optional[
        typing.Union[dict[str, typing.Any], NotSetType]
    ] = NotSet
    required_inputs: typing.Union[list[str], NotSetType] = NotSet
    timeout: typing.Optional[typing.Union[int, NotSetType]] = NotSet

    model_config = ConfigDict(
        extra="forbid", json_schema_extra=NotSetType.openapi_schema_modifier
    )



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import argparse
from operator import attrgetter
import typing


class DeprecationAction(argparse.Action):
    """An action that triggers a deprecation error."""

    __deprecation_msg: typing.Optional[str] = None

    def __init__(
        self,
        option_strings,
        dest,
        help=None,
        deprecation_msg: typing.Optional[str] = None,
        **kwargs,
    ):
        super(DeprecationAction, self).__init__(
            option_strings, dest, nargs=0, help=help, **kwargs
        )
        self.__deprecation_msg = deprecation_msg

    def __call__(self, parser, namespace, values, option_string=None):
        deprecation_msg = (
            self.__deprecation_msg
            if self.__deprecation_msg
            else f"Deprecated option: {option_string}"
        )
        parser.error(deprecation_msg)


class SortingHelpFormatter(argparse.HelpFormatter):
    def add_arguments(self, actions):
        actions = sorted(actions, key=attrgetter("option_strings"))
        super(SortingHelpFormatter, self).add_arguments(actions)



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import argparse

from ...domain.datasets import Dataset
from ..command import RobotoCommand
from ..context import CLIContext
from .shared_helpdoc import DATASET_ID_HELP


def list_files(args, context: CLIContext, parser: argparse.ArgumentParser):
    dataset = Dataset.from_id(args.dataset_id, context.roboto_client)

    for f in dataset.list_files(args.include, args.exclude):
        print(f"{f.relative_path}")


def list_files_setup_parser(parser):
    parser.add_argument(
        "-d", "--dataset-id", type=str, required=True, help=DATASET_ID_HELP
    )
    parser.add_argument(
        "-i",
        "--include",
        type=str,
        nargs="*",
        help="Zero or more include filters (if path points to a directory)",
    )
    parser.add_argument(
        "-x",
        "--exclude",
        type=str,
        nargs="*",
        help="Zero or more exclude filters (if path points to a directory)",
    )


list_files_command = RobotoCommand(
    name="list-files",
    logic=list_files,
    setup_parser=list_files_setup_parser,
    command_kwargs={"help": "Lists files for a specific dataset."},
)



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

from .action_runtime import (
    ActionRuntime,
    FilesChangesetFileManager,
)
from .association import (
    Association,
    AssociationType,
)
from .config import RobotoConfig
from .domain.actions import (
    Accessibility,
    Action,
    ActionParameter,
    ActionParameterChangeset,
    ActionProvenance,
    ActionRecord,
    ActionReference,
    ComputeRequirements,
    ContainerParameters,
    CreateActionRequest,
    CreateInvocationRequest,
    CreateTriggerRequest,
    EvaluateTriggersRequest,
    ExecutableProvenance,
    ExecutorContainer,
    Invocation,
    InvocationDataSource,
    InvocationDataSourceType,
    InvocationProvenance,
    InvocationRecord,
    InvocationSource,
    InvocationStatus,
    InvocationStatusRecord,
    LogRecord,
    LogsLocation,
    QueryTriggersRequest,
    SetActionAccessibilityRequest,
    SetContainerInfoRequest,
    SetLogsLocationRequest,
    SourceProvenance,
    Trigger,
    TriggerEvaluationCause,
    TriggerEvaluationOutcome,
    TriggerEvaluationOutcomeReason,
    TriggerEvaluationRecord,
    TriggerEvaluationsSummaryResponse,
    TriggerEvaluationStatus,
    TriggerForEachPrimitive,
    TriggerRecord,
    UpdateActionRequest,
    UpdateInvocationStatus,
    UpdateTriggerRequest,
)
from .domain.collections import (
    Collection,
    CollectionChangeRecord,
    CollectionChangeSet,
    CollectionContentMode,
    CollectionRecord,
    CollectionResourceRef,
    CollectionResourceType,
    CreateCollectionRequest,
    UpdateCollectionRequest,
)
from .domain.comments import (
    Comment,
    CommentEntityType,
    CommentRecord,
    CreateCommentRequest,
    UpdateCommentRequest,
)
from .domain.datasets import (
    BeginManifestTransactionRequest,
    BeginSingleFileUploadRequest,
    CreateDatasetRequest,
    Dataset,
    DatasetCredentials,
    DatasetRecord,
    QueryDatasetFilesRequest,
    QueryDatasetsRequest,
    ReportTransactionProgressRequest,
    UpdateDatasetRequest,
)
from .domain.devices import (
    CreateDeviceRequest,
    Device,
    DeviceRecord,
)
from .domain.events import Event, EventRecord
from .domain.files import (
    CredentialProvider,
    DeleteFileRequest,
    File,
    FileRecord,
    FileRecordRequest,
    FileStatus,
    FileTag,
    ImportFileRequest,
    IngestionStatus,
    QueryFilesRequest,
    S3Credentials,
    UpdateFileRecordRequest,
)
from .domain.topics import (
    AddMessagePathRepresentationRequest,
    AddMessagePathRequest,
    CanonicalDataType,
    CreateTopicRequest,
    DeleteMessagePathRequest,
    MessagePath,
    MessagePathChangeset,
    MessagePathRecord,
    MessagePathStatistic,
    RepresentationRecord,
    RepresentationStorageFormat,
    SetDefaultRepresentationRequest,
    Topic,
    TopicRecord,
    UpdateMessagePathRequest,
    UpdateTopicRequest,
)
from .domain.users import (
    CreateUserRequest,
    UpdateUserRequest,
    User,
    UserRecord,
)
from .env import RobotoEnv
from .http import BatchRequest, RobotoClient
from .regionalization import RobotoRegion
from .roboto_search import RobotoSearch
from .warnings import (
    roboto_default_warning_behavior,
)

__all__ = [
    "Accessibility",
    "Action",
    "ActionParameter",
    "ActionParameterChangeset",
    "ActionProvenance",
    "ActionRecord",
    "ActionReference",
    "ActionRuntime",
    "AddMessagePathRepresentationRequest",
    "AddMessagePathRequest",
    "Association",
    "AssociationType",
    "BatchRequest",
    "BeginManifestTransactionRequest",
    "BeginSingleFileUploadRequest",
    "CanonicalDataType",
    "Collection",
    "CollectionChangeRecord",
    "CollectionChangeSet",
    "CollectionContentMode",
    "CollectionRecord",
    "CollectionResourceRef",
    "CollectionResourceType",
    "Comment",
    "CommentEntityType",
    "CommentRecord",
    "ComputeRequirements",
    "ContainerParameters",
    "CreateActionRequest",
    "CreateCollectionRequest",
    "CreateCommentRequest",
    "CreateDatasetRequest",
    "CreateDeviceRequest",
    "CreateInvocationRequest",
    "CreateTopicRequest",
    "CreateTriggerRequest",
    "CreateUserRequest",
    "CredentialProvider",
    "Dataset",
    "DatasetCredentials",
    "DatasetRecord",
    "DeleteFileRequest",
    "DeleteMessagePathRequest",
    "Device",
    "DeviceRecord",
    "EvaluateTriggersRequest",
    "ExecutableProvenance",
    "ExecutorContainer",
    "Event",
    "EventRecord",
    "File",
    "FilesChangesetFileManager",
    "FileRecord",
    "FileRecordRequest",
    "FileStatus",
    "FileTag",
    "ImportFileRequest",
    "IngestionStatus",
    "Invocation",
    "InvocationDataSource",
    "InvocationDataSourceType",
    "InvocationProvenance",
    "InvocationRecord",
    "InvocationSource",
    "InvocationStatus",
    "InvocationStatusRecord",
    "LogRecord",
    "LogsLocation",
    "MessagePath",
    "MessagePathChangeset",
    "MessagePathRecord",
    "MessagePathStatistic",
    "QueryDatasetFilesRequest",
    "QueryDatasetsRequest",
    "QueryFilesRequest",
    "QueryTriggersRequest",
    "ReportTransactionProgressRequest",
    "RepresentationRecord",
    "RepresentationStorageFormat",
    "RobotoClient",
    "RobotoConfig",
    "RobotoEnv",
    "RobotoRegion",
    "RobotoSearch",
    "S3Credentials",
    "SetActionAccessibilityRequest",
    "SetContainerInfoRequest",
    "SetDefaultRepresentationRequest",
    "SetLogsLocationRequest",
    "SourceProvenance",
    "Topic",
    "TopicRecord",
    "Trigger",
    "TriggerEvaluationCause",
    "TriggerEvaluationOutcome",
    "TriggerEvaluationOutcomeReason",
    "TriggerEvaluationRecord",
    "TriggerEvaluationStatus",
    "TriggerEvaluationsSummaryResponse",
    "TriggerForEachPrimitive",
    "TriggerRecord",
    "UpdateActionRequest",
    "UpdateCollectionRequest",
    "UpdateCommentRequest",
    "UpdateDatasetRequest",
    "UpdateFileRecordRequest",
    "UpdateInvocationStatus",
    "UpdateMessagePathRequest",
    "UpdateTopicRequest",
    "UpdateTriggerRequest",
    "UpdateUserRequest",
    "User",
    "UserRecord",
]


roboto_default_warning_behavior()



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import collections.abc
import math
import typing

from .domain import (
    collections as roboto_collections,
)
from .domain import (
    datasets,
    events,
    files,
    topics,
)
from .query import Query, QueryClient, QueryTarget


class RobotoSearch:
    """
    A high-level interface for querying the Roboto data platform.

    For most customers, using this class should be as simple as:
        >>> from roboto import RobotoSearch
        >>> robosearch = RobotoSearch()
        >>> for dataset in robosearch.find_datasets(...):
        ...     ...
    """

    __query_client: QueryClient

    def __init__(self, query_client: typing.Optional[QueryClient] = None):
        self.__query_client = (
            query_client if query_client is not None else QueryClient()
        )

    def find_collections(
        self,
        query: typing.Optional[Query] = None,
        timeout_seconds: float = math.inf,
    ) -> collections.abc.Generator[roboto_collections.Collection, None, None]:
        for result in self.__query_client.submit_query(
            query, QueryTarget.Collections, timeout_seconds
        ):
            yield roboto_collections.Collection(
                roboto_collections.CollectionRecord.model_validate(result),
                self.__query_client.roboto_client,
            )

    def find_datasets(
        self,
        query: typing.Optional[Query] = None,
        timeout_seconds: float = math.inf,
    ) -> collections.abc.Generator[datasets.Dataset, None, None]:
        for result in self.__query_client.submit_query(
            query, QueryTarget.Datasets, timeout_seconds
        ):
            yield datasets.Dataset(
                datasets.DatasetRecord.model_validate(result),
                self.__query_client.roboto_client,
            )

    def find_files(
        self,
        query: typing.Optional[Query] = None,
        timeout_seconds: float = math.inf,
    ) -> collections.abc.Generator[files.File, None, None]:
        for result in self.__query_client.submit_query(
            query, QueryTarget.Files, timeout_seconds
        ):
            yield files.File(
                files.FileRecord.model_validate(result),
                self.__query_client.roboto_client,
            )

    def find_message_paths(
        self,
        query: typing.Optional[Query] = None,
        timeout_seconds: float = math.inf,
    ) -> collections.abc.Generator[topics.MessagePath, None, None]:
        for result in self.__query_client.submit_query(
            query, QueryTarget.TopicMessagePaths, timeout_seconds
        ):
            yield topics.MessagePath(
                topics.MessagePathRecord.model_validate(result),
                self.__query_client.roboto_client,
            )

    def find_topics(
        self,
        query: typing.Optional[Query] = None,
        timeout_seconds: float = math.inf,
    ) -> collections.abc.Generator[topics.Topic, None, None]:
        """
        Examples:
            >>> import matplotlib.pyplot as plt
            >>> import pandas as pd
            >>> from roboto import RobotoSearch
            >>> robosearch = RobotoSearch()
            >>> for topic in robosearch.find_topics("msgpaths[cpuload.load].max > 0.9"):
            ...     topic_data = list(topic.get_data())
            ...     df = pd.json_normalize(topic_data)
            ... plt.plot(df["log_time"], df["load"], label=f"{topic.topic_id}")
            ...
            >>> plt.legend()
            >>> plt.show()
        """
        for result in self.__query_client.submit_query(
            query, QueryTarget.Topics, timeout_seconds
        ):
            yield topics.Topic(
                topics.TopicRecord.model_validate(result),
                self.__query_client.roboto_client,
            )

    def find_events(
        self, query: typing.Optional[Query] = None, timeout_seconds: float = math.inf
    ) -> collections.abc.Generator[events.Event]:
        for result in self.__query_client.submit_query(
            query, QueryTarget.Events, timeout_seconds
        ):
            yield events.Event(
                events.EventRecord.model_validate(result),
                self.__query_client.roboto_client,
            )



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import argparse

from ...domain.collections import (
    Collection,
    CollectionContentMode,
)
from ..command import RobotoCommand
from ..common_args import add_org_arg
from ..context import CLIContext
from .shared_helpdoc import CONTENT_MODE_HELP


def list_all(args, context: CLIContext, parser: argparse.ArgumentParser):
    for collection in Collection.list_all(
        owner_org_id=args.org,
        roboto_client=context.roboto_client,
        content_mode=args.content_mode,
    ):
        print(collection.record.model_dump_json(indent=2))


def list_all_setup_parser(parser: argparse.ArgumentParser):
    parser.add_argument(
        "--content-mode",
        type=str,
        choices=[mode.value for mode in CollectionContentMode],
        help=CONTENT_MODE_HELP,
        default=CollectionContentMode.References.value,
    )
    add_org_arg(parser)


list_all_command = RobotoCommand(
    name="list-all",
    logic=list_all,
    setup_parser=list_all_setup_parser,
    command_kwargs={"help": "Lists all collections created for a given org."},
)



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import argparse

from ...domain import actions
from ..command import (
    KeyValuePairsAction,
    RobotoCommand,
)
from ..common_args import (
    ActionTimeoutArg,
    add_action_reference_arg,
    add_compute_requirements_args,
    add_container_parameters_args,
    add_org_arg,
    parse_compute_requirements,
    parse_container_overrides,
)
from ..context import CLIContext


def invoke(
    args: argparse.Namespace, context: CLIContext, parser: argparse.ArgumentParser
) -> None:
    owner_org_id = args.action.owner if args.action.owner else args.org
    action = actions.Action.from_name(
        name=args.action.name,
        digest=args.action.digest,
        owner_org_id=owner_org_id,
        roboto_client=context.roboto_client,
    )

    compute_requirements = parse_compute_requirements(args, action.compute_requirements)
    container_parameters = parse_container_overrides(args, action.container_parameters)

    invocation = action.invoke(
        input_data=args.input_data,
        data_source_id=args.dataset_id,
        data_source_type=actions.InvocationDataSourceType.Dataset,
        invocation_source=actions.InvocationSource.Manual,
        parameter_values=args.parameter_value,
        compute_requirement_overrides=compute_requirements,
        container_parameter_overrides=container_parameters,
        idempotency_id=args.idempotency_id,
        timeout=args.timeout,
        caller_org_id=args.org,
    )
    print(
        f"Queued invocation of '{action.record.reference!r}'. Invocation ID: '{invocation.id}'"
    )


def invoke_parser(parser: argparse.ArgumentParser) -> None:
    add_action_reference_arg(parser)
    add_org_arg(parser)
    parser.add_argument(
        "--input-data",
        required=True,
        dest="input_data",
        type=str,
        nargs="+",
        action="extend",
        help=(
            "One or many file patterns for data to download from the data source. Examples: "
            "front camera images, ``--input-data '**/cam_front/*.jpg'``; "
            "front and rear camera images, ``--input-data '**/cam_front/*.jpg' --input-data '**/cam_rear/*.jpg'``; "
            "all data, ``--input-data '**/*'``."
        ),
    )

    parser.add_argument(
        "--timeout",
        required=False,
        action=ActionTimeoutArg,
        help="Optional timeout for an action in minutes, defaults to 30 minutes or 12 hours depending on tier.",
    )

    parser.add_argument(
        "--idempotency_id",
        dest="idempotency_id",
        type=str,
        help="Optional unique ID which ensures that an invocation is run exactly once.",
    )

    # As more InvocationDataSourceTypes are supported,
    # additional, mutually-exclusive argument groups should be added
    dataset_source_group = parser.add_argument_group(
        "Dataset",
        "Use a dataset as the data source for this invocation. This is currently the only option.",
    )
    dataset_source_group.add_argument(
        "--dataset-id",
        required=True,
        action="store",
        dest="dataset_id",
        help="Unique identifier for dataset to use as data source for this invocation.",
    )

    parser.add_argument(
        "--parameter-value",
        required=False,
        metavar="<PARAMETER_NAME>=<PARAMETER_VALUE>",
        nargs="*",
        action=KeyValuePairsAction,
        help=(
            "Zero or more ``<parameter_name>=<parameter_value>`` pairs. "
            "``parameter_value`` is parsed as JSON. "
        ),
    )

    add_compute_requirements_args(parser)
    add_container_parameters_args(parser)


invoke_command = RobotoCommand(
    name="invoke",
    logic=invoke,
    setup_parser=invoke_parser,
    command_kwargs={"help": "Invoke an action by name."},
)



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import collections.abc
import datetime
import decimal
import enum
import json
import typing

import pydantic

from ..collection_utils import get_by_path

ConditionValue = typing.Optional[
    typing.Union[str, bool, int, float, decimal.Decimal, datetime.datetime]
]


class Comparator(str, enum.Enum):
    """The comparator to use when comparing a field to a value."""

    Equals = "EQUALS"
    NotEquals = "NOT_EQUALS"
    GreaterThan = "GREATER_THAN"
    GreaterThanOrEqual = "GREATER_THAN_OR_EQUAL"
    LessThan = "LESS_THAN"
    LessThanOrEqual = "LESS_THAN_OR_EQUAL"
    Contains = "CONTAINS"
    NotContains = "NOT_CONTAINS"
    IsNull = "IS_NULL"
    IsNotNull = "IS_NOT_NULL"
    Exists = "EXISTS"
    NotExists = "NOT_EXISTS"
    BeginsWith = "BEGINS_WITH"
    Like = "LIKE"  # SQL Syntax
    NotLike = "NOT_LIKE"  # SQL Syntax

    def to_compact_string(self):
        if self is Comparator.Equals:
            return "="
        elif self is Comparator.NotEquals:
            return "!="
        elif self is Comparator.GreaterThanOrEqual:
            return ">"
        elif self is Comparator.GreaterThan:
            return ">="
        elif self is Comparator.LessThan:
            return "<"
        elif self is Comparator.LessThanOrEqual:
            return "<="
        else:
            return self.value

    @staticmethod
    def from_string(value: str) -> "Comparator":
        as_comparator: typing.Optional[Comparator] = {
            "EQUALS": Comparator.Equals,
            "=": Comparator.Equals,
            "NOT_EQUALS": Comparator.NotEquals,
            "!=": Comparator.NotEquals,
            "GREATER_THAN": Comparator.GreaterThan,
            ">": Comparator.GreaterThan,
            "GREATER_THAN_OR_EQUAL": Comparator.GreaterThanOrEqual,
            ">=": Comparator.GreaterThanOrEqual,
            "LESS_THAN": Comparator.LessThan,
            "<": Comparator.LessThan,
            "LESS_THAN_OR_EQUAL": Comparator.LessThanOrEqual,
            "<=": Comparator.LessThanOrEqual,
            "CONTAINS": Comparator.Contains,
            "NOT_CONTAINS": Comparator.NotContains,
            "IS_NULL": Comparator.IsNull,
            "IS_NOT_NULL": Comparator.IsNotNull,
            "EXISTS": Comparator.Exists,
            "NOT_EXISTS": Comparator.NotExists,
            "BEGINS_WITH": Comparator.BeginsWith,
            "LIKE": Comparator.Like,
            "NOT_LIKE": Comparator.NotLike,
        }.get(value.upper())

        if as_comparator is None:
            raise ValueError(f"Unrecognized comparator {value}")

        return as_comparator


class Condition(pydantic.BaseModel):
    """A filter for any arbitrary attribute for a Roboto resource."""

    field: str
    comparator: Comparator
    value: ConditionValue = None

    def matches(self, target: dict) -> bool:
        value = get_by_path(target, self.field.split("."))

        if self.comparator in [Comparator.NotExists, Comparator.IsNull]:
            return value is None

        if self.comparator in [Comparator.Exists, Comparator.IsNotNull]:
            return value is not None

        # We need the value for everything else
        if value is None:
            return False

        if isinstance(value, str) and not isinstance(self.value, str):
            if isinstance(self.value, int):
                value = int(value)
            elif isinstance(self.value, float):
                value = float(value)
            elif isinstance(self.value, bool):
                value = value.lower() == "true"
            elif isinstance(self.value, decimal.Decimal):
                value = decimal.Decimal.from_float(float(value))
            elif isinstance(self.value, datetime.datetime):
                value = datetime.datetime.fromisoformat(value)

        if self.comparator is Comparator.Equals:
            return value == self.value

        if self.comparator is Comparator.NotEquals:
            return value != self.value

        if self.comparator is Comparator.GreaterThan:
            return value > self.value

        if self.comparator is Comparator.GreaterThanOrEqual:
            return value >= self.value

        if self.comparator is Comparator.LessThan:
            return value < self.value

        if self.comparator is Comparator.LessThanOrEqual:
            return value <= self.value

        if self.comparator is Comparator.Contains:
            return isinstance(value, list) and self.value in value

        if self.comparator is Comparator.NotContains:
            return isinstance(value, list) and self.value not in value

        if self.comparator is Comparator.BeginsWith:
            if not isinstance(value, str) or not isinstance(self.value, str):
                return False

            return value.startswith(self.value)

        return False

    def __str__(self):
        base_condition = f"{self.field} {self.comparator.to_compact_string()}"
        if self.value is None:
            return base_condition

        if isinstance(self.value, datetime.datetime):
            return f"{base_condition} {self.value.isoformat()}"

        return f"{base_condition} {json.dumps(self.value)}"

    def __repr__(self):
        return self.model_dump_json()

    @classmethod
    def equals_cond(cls, field: str, value: ConditionValue) -> "Condition":
        return cls(field=field, comparator=Comparator.Equals, value=value)


class ConditionOperator(str, enum.Enum):
    """The operator to use when combining multiple conditions."""

    And = "AND"
    Or = "OR"
    Not = "NOT"

    @staticmethod
    def from_string(value: str) -> "ConditionOperator":
        if value.lower() == "and":
            return ConditionOperator.And
        elif value.lower() == "or":
            return ConditionOperator.Or
        elif value.lower() == "not":
            return ConditionOperator.Not

        raise ValueError(f"Unrecognized Operator {value}")


class ConditionGroup(pydantic.BaseModel):
    """A group of conditions that are combined together."""

    operator: ConditionOperator
    conditions: collections.abc.Sequence[typing.Union[Condition, "ConditionGroup"]]

    def matches(self, target: dict):
        inner_matches = map(lambda x: x.matches(target), self.conditions)

        if self.operator is ConditionOperator.And:
            return all(inner_matches)

        if self.operator is ConditionOperator.Or:
            return any(inner_matches)

        if self.operator is ConditionOperator.Not:
            return not any(inner_matches)

        return False

    @staticmethod
    def or_group(
        *conditions: typing.Union[Condition, "ConditionGroup"]
    ) -> "ConditionGroup":
        return ConditionGroup(operator=ConditionOperator.Or, conditions=conditions)

    @staticmethod
    def and_group(
        *conditions: typing.Union[Condition, "ConditionGroup"]
    ) -> "ConditionGroup":
        return ConditionGroup(operator=ConditionOperator.And, conditions=conditions)

    @pydantic.field_validator("conditions")
    def validate_conditions(
        cls, v: collections.abc.Sequence[typing.Union[Condition, "ConditionGroup"]]
    ):
        if len(v) == 0:
            raise ValueError(
                "At least one condition must be provided to a ConditionGroup, got 0!"
            )

        return v

    def __str__(self):
        return (
            "("
            + f" {self.operator.value} ".join(
                [str(condition) for condition in self.conditions]
            )
            + ")"
        )

    def __repr__(self):
        return self.model_dump_json()


ConditionType = typing.Union[Condition, ConditionGroup]



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import collections.abc
import math
import typing

import mcap.reader
import mcap_ros1.decoder
import mcap_ros2.decoder

from .decoded_message import DecodedMessage
from .json_decoder_factory import (
    JsonDecoderFactory,
)
from .record import MessagePathRecord


class McapReader:
    __message_iterator: typing.Iterator[mcap.reader.DecodedMessageTuple]
    __message_paths: collections.abc.Sequence[MessagePathRecord]
    __next_unconsummed_decode_result: mcap.reader.DecodedMessageTuple | None = None

    def __init__(
        self,
        stream: typing.IO[bytes],
        message_paths: collections.abc.Sequence[MessagePathRecord],
        start_time: typing.Optional[int] = None,
        end_time: typing.Optional[int] = None,
    ):
        json_decoder = JsonDecoderFactory()
        ros1_decoder = mcap_ros1.decoder.DecoderFactory()
        ros2_decoder = mcap_ros2.decoder.DecoderFactory()
        reader = mcap.reader.make_reader(
            stream, decoder_factories=[json_decoder, ros1_decoder, ros2_decoder]
        )
        self.__message_iterator = reader.iter_decoded_messages(
            start_time=start_time, end_time=end_time
        )
        self.__message_paths = message_paths
        self.__decode_next()

    @property
    def has_next(self) -> bool:
        return self.__next_unconsummed_decode_result is not None

    @property
    def next_timestamp(self) -> typing.Union[int, float]:
        if self.__next_unconsummed_decode_result is None:
            return math.inf

        return self.__next_unconsummed_decode_result.message.log_time

    def next(self) -> DecodedMessage | None:
        next_decode_result = self.__next_unconsummed_decode_result
        self.__decode_next()

        if next_decode_result is not None:
            return DecodedMessage(
                next_decode_result.decoded_message, self.__message_paths
            )
        return None

    def next_message_is_time_aligned(self, timestamp: typing.Union[int, float]) -> bool:
        next_decode_result = self.__next_unconsummed_decode_result
        if next_decode_result is None:
            return False

        return next_decode_result.message.log_time == timestamp

    def __decode_next(self) -> None:
        next_decode_result = next(self.__message_iterator, None)
        self.__next_unconsummed_decode_result = next_decode_result



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import enum


class Permissions(enum.Enum):
    """
    Enum for permission levels of a Roboto resource.
    It is a best practice to only request/use the minimum permissions required for a given operation.

    For example:
    - When listing files associated with a dataset or pulling a container image hosted in Roboto's registry,
      use `ReadOnly` permissions.
    - When adding files to a dataset or pushing a container image to Roboto's registry, use `ReadWrite` permissions.
    """

    ReadOnly = "ReadOnly"
    ReadWrite = "ReadWrite"



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

from .constants import (
    BEARER_TOKEN_HEADER,
    ORG_OVERRIDE_HEADER,
    ORG_OVERRIDE_QUERY_PARAM,
    RESOURCE_OWNER_OVERRIDE_HEADER,
    RESOURCE_OWNER_OVERRIDE_QUERY_PARAM,
    USER_OVERRIDE_HEADER,
    USER_OVERRIDE_QUERY_PARAM,
)
from .headers import (
    CONTENT_TYPE_JSON_HEADER,
    roboto_headers,
)
from .http_client import (
    ClientError,
    HttpClient,
    HttpError,
    ServerError,
)
from .request import BatchRequest
from .request_decorators import (
    BearerTokenDecorator,
    SigV4AuthDecorator,
)
from .requester import (
    ROBOTO_REQUESTER_HEADER,
    RobotoRequester,
    RobotoTool,
)
from .response import (
    BatchResponse,
    BatchResponseElement,
    PaginatedList,
    PaginationToken,
    PaginationTokenEncoding,
    PaginationTokenScheme,
    StreamedList,
)
from .roboto_client import RobotoClient
from .testing_util import FakeHttpResponseFactory

__all__ = (
    "BEARER_TOKEN_HEADER",
    "BatchRequest",
    "BatchResponse",
    "BatchResponseElement",
    "BearerTokenDecorator",
    "CONTENT_TYPE_JSON_HEADER",
    "ClientError",
    "FakeHttpResponseFactory",
    "HttpClient",
    "HttpError",
    "ORG_OVERRIDE_HEADER",
    "ORG_OVERRIDE_QUERY_PARAM",
    "PaginatedList",
    "PaginationToken",
    "PaginationTokenEncoding",
    "PaginationTokenScheme",
    "RESOURCE_OWNER_OVERRIDE_HEADER",
    "RESOURCE_OWNER_OVERRIDE_QUERY_PARAM",
    "ROBOTO_REQUESTER_HEADER",
    "ServerError",
    "SigV4AuthDecorator",
    "StreamedList",
    "USER_OVERRIDE_HEADER",
    "USER_OVERRIDE_QUERY_PARAM",
    "roboto_headers",
    "RobotoClient",
    "RobotoRequester",
    "RobotoTool",
)



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import argparse
import json

from ...domain import datasets
from ..command import (
    KeyValuePairsAction,
    RobotoCommand,
)
from ..common_args import add_org_arg
from ..context import CLIContext


def create(args, context: CLIContext, parser: argparse.ArgumentParser):
    dataset = datasets.Dataset.create(
        description=args.description,
        metadata=args.metadata,
        tags=args.tag,
        roboto_client=context.roboto_client,
        caller_org_id=args.org,
    )

    print(json.dumps(dataset.to_dict(), indent=2))


def create_setup_parser(parser):
    parser.add_argument(
        "-m",
        "--metadata",
        metavar="KEY=VALUE",
        nargs="*",
        action=KeyValuePairsAction,
        help="Zero or more ``<key>=<value>`` pairs which represent dataset metadata. "
        + "Metadata can be changed after creation.",
    )
    parser.add_argument(
        "-t",
        "--tag",
        type=str,
        nargs="*",
        help="One or more tags to annotate this dataset. Tags can be modified after creation.",
        action="extend",
    )
    parser.add_argument(
        "-d",
        "--description",
        type=str,
        help="A human readable description of this dataset.",
    )
    add_org_arg(parser=parser)


create_command = RobotoCommand(
    name="create",
    logic=create,
    setup_parser=create_setup_parser,
    command_kwargs={"help": "Creates a new dataset."},
)



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import warnings


def roboto_default_warning_behavior():
    # https://github.com/boto/botocore/issues/619
    warnings.filterwarnings(
        "ignore",
        module="botocore.vendored.requests.packages.urllib3.connectionpool",
        message=".*",
    )



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import argparse
import json

from ...domain.datasets import Dataset
from ..command import RobotoCommand
from ..context import CLIContext
from .shared_helpdoc import DATASET_ID_HELP


def show(args, context: CLIContext, parser: argparse.ArgumentParser):
    record = Dataset.from_id(args.dataset_id, context.roboto_client)
    print(json.dumps(record.to_dict(), indent=2))


def show_setup_parser(parser):
    parser.add_argument(
        "-d", "--dataset-id", type=str, required=True, help=DATASET_ID_HELP
    )


show_command = RobotoCommand(
    name="show",
    logic=show,
    setup_parser=show_setup_parser,
    command_kwargs={"help": "Show information about a specific dataset."},
)



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import collections.abc
import typing

from ...exceptions import RobotoDomainException
from ...http import RobotoClient
from .operations import CreateTokenRequest
from .record import TokenRecord


class Token:
    """
    A token allows users and devices to authenticate requests to Roboto.

    When a token is generated, a secret value is created which is shared with the user exactly once. This value is
    required for authentication, if lost a token can be deleted and a new token can be generated.
    """

    __roboto_client: RobotoClient
    __record: TokenRecord

    @classmethod
    def create(
        cls,
        name: str,
        description: typing.Optional[str] = None,
        expiry_days: int = 30,
        roboto_client: typing.Optional[RobotoClient] = None,
    ) -> tuple["Token", str]:
        """
        This returns a tuple of the token and the one-time generated secret. After creation, you'll never be able
        to retrieve the secret again, so you should save it!
        """
        roboto_client = RobotoClient.defaulted(roboto_client)

        request = CreateTokenRequest(
            name=name, description=description, expiry_days=expiry_days
        )

        record = roboto_client.post("v1/tokens", data=request).to_record(TokenRecord)

        if record.secret is None:
            raise RobotoDomainException(
                "Token was generated without returning secret value, this renders it unusable."
            )

        return cls(record=record, roboto_client=roboto_client), record.secret

    @classmethod
    def from_id(
        cls, token_id: str, roboto_client: typing.Optional[RobotoClient] = None
    ) -> "Token":
        roboto_client = RobotoClient.defaulted(roboto_client)
        record = roboto_client.get(f"v1/tokens/id/{token_id}").to_record(TokenRecord)
        return cls(record=record, roboto_client=roboto_client)

    @classmethod
    def for_self(
        cls, roboto_client: typing.Optional[RobotoClient] = None
    ) -> collections.abc.Collection["Token"]:
        roboto_client = RobotoClient.defaulted(roboto_client)
        records = roboto_client.get("v1/tokens").to_record_list(TokenRecord)

        return [cls(record=record, roboto_client=roboto_client) for record in records]

    def __init__(
        self, record: TokenRecord, roboto_client: typing.Optional[RobotoClient] = None
    ):
        self.__record = record
        self.__roboto_client = RobotoClient.defaulted(roboto_client)

    def __repr__(self) -> str:
        return self.__record.model_dump_json()

    @property
    def record(self) -> TokenRecord:
        return self.__record

    @property
    def token_id(self) -> str:
        if self.__record.context is None:
            raise ValueError("Token record did not contain a token_id: " + str(self))

        return self.__record.context.token_id

    @property
    def user_id(self) -> str:
        if self.__record.user_id is None:
            raise ValueError("Token record did not contain a user_id: " + str(self))

        return self.__record.user_id

    def delete(self) -> None:
        self.__roboto_client.delete(f"v1/tokens/id/{self.token_id}")

    def disable(self) -> "Token":
        self.__roboto_client.post(f"v1/tokens/id/{self.token_id}/disable")
        return self

    def enable(self) -> "Token":
        self.__roboto_client.post(f"v1/tokens/id/{self.token_id}/enable")
        return self

    def to_dict(self) -> dict[str, typing.Any]:
        return self.__record.model_dump(mode="json")



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

from typing import Optional

from .conditions import (
    Comparator,
    Condition,
    ConditionGroup,
    ConditionOperator,
    ConditionType,
)


def __pathspec_to_like_value(pattern: str) -> str:
    return pattern.replace("**/*", "%").replace("**", "%").replace("*", "%")


def git_paths_to_condition_group(
    include_patterns: Optional[list[str]],
    exclude_patterns: Optional[list[str]],
    path_field_name: str,
) -> ConditionGroup:
    conditions: list[ConditionType] = []

    if include_patterns:
        include_pattern_conditions = list(
            map(
                lambda pattern: Condition(
                    field=path_field_name,
                    comparator=Comparator.Like,
                    value=__pathspec_to_like_value(pattern),
                ),
                include_patterns,
            )
        )

        conditions.append(
            ConditionGroup(
                conditions=include_pattern_conditions, operator=ConditionOperator.Or
            )
        )

    if exclude_patterns:
        exclude_pattern_conditions = list(
            map(
                lambda pattern: Condition(
                    field=path_field_name,
                    comparator=Comparator.NotLike,
                    value=__pathspec_to_like_value(pattern),
                ),
                exclude_patterns,
            )
        )

        conditions.append(
            ConditionGroup(
                conditions=exclude_pattern_conditions, operator=ConditionOperator.And
            )
        )

    return ConditionGroup(conditions=conditions, operator=ConditionOperator.And)



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

from __future__ import annotations

import collections.abc
import dataclasses
import typing

from ...domain import events
from ...http import RobotoClient

if typing.TYPE_CHECKING:
    import pandas  # pants: no-infer-dep


@dataclasses.dataclass(frozen=True)
class MatchContext:
    """
    Correlate a matched subsequence back to its source.
    """

    message_paths: collections.abc.Sequence[str]
    topic_id: str
    topic_name: str

    dataset_id: typing.Optional[str] = None
    file_id: typing.Optional[str] = None


@dataclasses.dataclass(frozen=True)
class Match:
    """
    A subsequence of a target signal that is similar to a query signal.
    """

    context: MatchContext
    """
    Correlate a matched subsequence back to its source.
    """

    end_idx: int
    """
    The end index in the target signal of this match.
    """

    end_time: int
    """
    The end time in the target signal of this match.
    """

    distance: float
    """
    Unitless measure of similarity between a query signal
    and the subsequence of the target signal this Match represents.
    A smaller distance relative to a larger distance indicates a "closer" match.
    """

    start_idx: int
    """
    The start index in the target signal of this match.
    """

    start_time: int
    """
    The start time in the target signal of this match.
    """

    subsequence: pandas.DataFrame
    """
    The subsequence of the target signal this Match represents.
    It is equivalent to ``target[start_idx:end_idx]``.
    """

    def to_event(
        self,
        name: str = "Signal Similarity Match Result",
        caller_org_id: typing.Optional[str] = None,
        roboto_client: typing.Optional[RobotoClient] = None,
    ) -> events.Event:
        """
        Create a Roboto Platform event out of this similarity match result.
        """
        return events.Event.create(
            description=f"Match score: {self.distance}",
            end_time=self.end_time,
            name=name,
            metadata={
                "distance": self.distance,
                "message_paths": self.context.message_paths,
                "start_index": self.start_idx,
                "end_index": self.end_idx,
            },
            start_time=self.start_time,
            topic_ids=[self.context.topic_id],
            caller_org_id=caller_org_id,
            roboto_client=roboto_client,
        )



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import argparse
import pathlib
import sys

from ..config import (
    RobotoConfig,
    RobotoConfigFileV1,
)
from ..domain import devices
from ..exceptions import RobotoConflictException
from ..logging import default_logger
from .command import (
    RobotoCommand,
    RobotoCommandSet,
)
from .common_args import (
    add_org_arg,
    get_defaulted_org_id,
)
from .context import CLIContext

logger = default_logger()


def __add_device_arg(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        "device_id",
        type=str,
        help="The device ID of the target device.",
    )


def delete_device(
    args: argparse.Namespace, context: CLIContext, parser: argparse.ArgumentParser
) -> None:
    org_id = get_defaulted_org_id(args.org)

    device = devices.Device.from_id(
        device_id=args.device_id,
        org_id=org_id,
        roboto_client=context.roboto_client,
    )

    device.delete()


def delete_device_setup_parser(parser: argparse.ArgumentParser) -> None:
    __add_device_arg(parser)
    add_org_arg(parser)


def disable_device_access(
    args: argparse.Namespace, context: CLIContext, parser: argparse.ArgumentParser
) -> None:
    org_id = get_defaulted_org_id(args.org)

    device = devices.Device.from_id(
        device_id=args.device_id,
        org_id=org_id,
        roboto_client=context.roboto_client,
    )

    for token in device.tokens():
        token.disable()


def disable_device_access_setup_parser(parser: argparse.ArgumentParser) -> None:
    __add_device_arg(parser)
    add_org_arg(parser)


def enable_device_access(
    args: argparse.Namespace, context: CLIContext, parser: argparse.ArgumentParser
) -> None:
    org_id = get_defaulted_org_id(args.org)

    device = devices.Device.from_id(
        device_id=args.device_id,
        org_id=org_id,
        roboto_client=context.roboto_client,
    )

    for token in device.tokens():
        token.enable()


def enable_device_access_setup_parser(parser: argparse.ArgumentParser) -> None:
    __add_device_arg(parser)
    add_org_arg(parser)


def generate_device_creds(
    args: argparse.Namespace, context: CLIContext, parser: argparse.ArgumentParser
) -> None:
    config_file: pathlib.Path = args.config_file
    if config_file.is_file():
        parser.error(
            f"Specified config file {args.config_file} already exists and would be overwritten. "
            + "Please provide a different --config-file, or delete the existing one."
        )

    org_id = get_defaulted_org_id(args.org)

    device = devices.Device.from_id(
        device_id=args.device_id,
        roboto_client=context.roboto_client,
        org_id=org_id,
    )

    generate_device_creds_common_logic(device, config_file, context)


def generate_device_creds_common_logic(
    device: devices.Device,
    config_file: pathlib.Path,
    context: CLIContext,
):
    """
    Used in `generate-creds` and `register`
    """
    device_token, device_secret = device.create_token()

    profiles: dict[str, RobotoConfig] = {}

    # Name the default profile after the fully qualified device ID
    default_profile = f"{device.org_id}_{device.device_id}"
    profiles[default_profile] = RobotoConfig(
        api_key=device_secret,
        endpoint=context.roboto_client.endpoint,
    )

    config_file_content = RobotoConfigFileV1(
        version="v1", profiles=profiles, default_profile=default_profile
    )

    config_file.write_text(config_file_content.model_dump_json())

    print(
        (
            f"Saved Roboto credentials for device {device.device_id} to {config_file.resolve()}\n\n"
            "This file should be copied to '$HOME/.roboto/config.json' on the device's operating system.\n\n"
            "You can also store this file in a different location, and use the environment variable "
            "ROBOTO_CONFIG_FILE to point to its location.\n\n"
            "Any first-party Roboto software, including `roboto` and "
            "`roboto-agent`, will use this file to authenticate with Roboto."
        ),
        file=sys.stderr,
    )


def generate_device_creds_setup_parser(parser: argparse.ArgumentParser) -> None:
    __add_device_arg(parser)
    add_org_arg(parser)
    parser.add_argument(
        "-f",
        "--config-file",
        type=pathlib.Path,
        default="config.json",
        help="Where on local disk to save the config file containing the device's generated Roboto credentials. "
        + "Defaults to 'config.json' in the current directory.",
    )


def list_devices(
    args: argparse.Namespace, context: CLIContext, parser: argparse.ArgumentParser
) -> None:
    org_id = get_defaulted_org_id(args.org)
    for device in devices.Device.for_org(org_id):
        print(device)


def list_devices_setup_parser(parser: argparse.ArgumentParser) -> None:
    add_org_arg(parser)


def register_device(
    args: argparse.Namespace, context: CLIContext, parser: argparse.ArgumentParser
) -> None:
    config_file: pathlib.Path = args.config_file
    if config_file.is_file():
        parser.error(
            f"Specified config file {args.config_file} already exists and would be overwritten. "
            + "Please provide a different --config-file, or delete the existing one."
        )

    org_id = get_defaulted_org_id(args.org)

    try:
        device = devices.Device.create(
            device_id=args.device_id,
            caller_org_id=org_id,
            roboto_client=context.roboto_client,
        )
    except RobotoConflictException:
        device = devices.Device.from_id(
            device_id=args.device_id,
            org_id=org_id,
            roboto_client=context.roboto_client,
        )

        print(
            f"Device {device.device_id} already exists, generating new credentials for it.\n",
            file=sys.stderr,
        )

    generate_device_creds_common_logic(device, config_file, context)


def register_device_setup_parser(parser: argparse.ArgumentParser) -> None:
    __add_device_arg(parser)
    parser.add_argument(
        "-f",
        "--config-file",
        type=pathlib.Path,
        default="config.json",
        help="Where on local disk to save the config file containing the newly registered device's Roboto credentials. "
        + "Defaults to 'config.json' in the current directory.",
    )
    add_org_arg(parser)


def show_device(
    args: argparse.Namespace, context: CLIContext, parser: argparse.ArgumentParser
) -> None:
    org_id = get_defaulted_org_id(args.org)

    print(
        devices.Device.from_id(
            device_id=args.device_id,
            roboto_client=context.roboto_client,
            org_id=org_id,
        )
    )


def show_device_setup_parser(parser: argparse.ArgumentParser) -> None:
    __add_device_arg(parser)
    add_org_arg(parser)


delete_command = RobotoCommand(
    name="delete",
    logic=delete_device,
    setup_parser=delete_device_setup_parser,
    command_kwargs={"help": "Deletes a device"},
)

disable_access_command = RobotoCommand(
    name="disable",
    logic=disable_device_access,
    setup_parser=disable_device_access_setup_parser,
    command_kwargs={
        "help": "Disables a device's access tokens without deleting them. "
        + "This is useful if a device is acting odd, and you want to temporarily "
        + "cut it off from uploading datasets, but you don't want to have to re-provision its Roboto access."
    },
)

enable_access_command = RobotoCommand(
    name="enable",
    logic=enable_device_access,
    setup_parser=enable_device_access_setup_parser,
    command_kwargs={
        "help": "Re-enables all of a device's access tokens. "
        + "They start enabled, so this is only needed if they've been explicitly disabled."
    },
)

generate_creds_command = RobotoCommand(
    name="generate-creds",
    logic=generate_device_creds,
    setup_parser=generate_device_creds_setup_parser,
    command_kwargs={
        "help": "Generates a new access credentials file for an existing device."
    },
)

list_command = RobotoCommand(
    name="list",
    logic=list_devices,
    setup_parser=list_devices_setup_parser,
    command_kwargs={
        "help": "Lists devices which have been registered for a given org."
    },
)

register_command = RobotoCommand(
    name="register",
    logic=register_device,
    setup_parser=register_device_setup_parser,
    command_kwargs={"help": "Registers a device with Roboto."},
)

show_command = RobotoCommand(
    name="show",
    logic=show_device,
    setup_parser=show_device_setup_parser,
    command_kwargs={"help": "Gets info about an individual device."},
)

commands = [
    enable_access_command,
    disable_access_command,
    delete_command,
    generate_creds_command,
    list_command,
    register_command,
    show_command,
]

command_set = RobotoCommandSet(
    name="devices",
    help="Manage the robots and other devices which have Roboto access within your org.",
    commands=commands,
)



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

from __future__ import annotations

import dataclasses
import typing

if typing.TYPE_CHECKING:
    import numpy.typing  # pants: no-infer-dep


@dataclasses.dataclass(frozen=True)
class Match:
    """
    A subsequence of a target signal that is similar to a query signal.
    """

    start_idx: int
    """
    The start index in the target signal of this match.
    """

    end_idx: int
    """
    The end index in the target signal of this match.
    """

    distance: float
    """
    Unitless measure of similarity between a query signal
    and the subsequence of the target signal this Match represents.
    A smaller distance relative to a larger distance indicates a "closer" match.
    """

    subsequence: numpy.typing.NDArray
    """
    The subsequence of the target signal this Match represents.
    It is equivalent to ``target_sequence[start_idx:end_idx]``.
    """



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import datetime
import enum
import typing

import pydantic


class CollectionResourceType(str, enum.Enum):
    """Type of resource added to a collection"""

    Dataset = "dataset"
    File = "file"


class CollectionContentMode(str, enum.Enum):
    """Desired content mode for representing a collection"""

    SummaryOnly = "summary_only"
    References = "references"
    Full = "full"


class CollectionResourceRef(pydantic.BaseModel):
    """Reference to a collection resource"""

    resource_type: CollectionResourceType
    resource_id: str
    resource_version: typing.Optional[str] = None


class CollectionRecord(pydantic.BaseModel):
    """A wire-transmissible representation of a collection"""

    collection_id: str
    name: typing.Optional[str] = None
    description: typing.Optional[str] = None
    resources: dict[CollectionResourceType, list[typing.Any]] = pydantic.Field(
        default_factory=dict
    )
    missing: dict[CollectionResourceType, list[CollectionResourceRef]] = pydantic.Field(
        default_factory=dict
    )
    tags: list[str] = []
    version: int
    created: datetime.datetime
    created_by: str
    updated: datetime.datetime
    updated_by: str
    org_id: str


class CollectionChangeSet(pydantic.BaseModel):
    """Changeset for updating a collection"""

    added_resources: list[CollectionResourceRef] = pydantic.Field(default_factory=list)
    added_tags: list[str] = pydantic.Field(default_factory=list)
    removed_resources: list[CollectionResourceRef] = pydantic.Field(
        default_factory=list
    )
    removed_tags: list[str] = pydantic.Field(default_factory=list)
    field_changes: dict[str, typing.Any] = pydantic.Field(default_factory=list)


class CollectionChangeRecord(pydantic.BaseModel):
    """A wire-transmissible representation of a collection change record"""

    collection_id: str
    from_version: int
    to_version: int
    change_set: CollectionChangeSet
    applied: datetime.datetime
    applied_by: str



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

from .comment import Comment
from .operations import (
    CreateCommentRequest,
    UpdateCommentRequest,
)
from .record import (
    CommentEntityType,
    CommentRecord,
)

__all__ = (
    "Comment",
    "CommentRecord",
    "CreateCommentRequest",
    "CommentEntityType",
    "UpdateCommentRequest",
)



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import abc
import typing

from .conditions import (
    Condition,
    ConditionGroup,
    ConditionType,
)


class ConditionVisitor(abc.ABC):
    """Query condition visitor"""

    def visit(self, cond: ConditionType) -> typing.Optional[ConditionType]:
        if isinstance(cond, Condition):
            return self.visit_condition(cond)
        elif isinstance(cond, ConditionGroup):
            return self.visit_condition_group(cond)
        else:
            raise ValueError(f"Unknown condition type: {type(cond)}")

    @abc.abstractmethod
    def visit_condition(self, condition: Condition) -> typing.Optional[ConditionType]:
        raise NotImplementedError("visit_condition")

    @abc.abstractmethod
    def visit_condition_group(
        self, condition_group: ConditionGroup
    ) -> typing.Optional[ConditionType]:
        raise NotImplementedError("visit_condition_group")


class BaseVisitor(ConditionVisitor):
    """Query base visitor"""

    def visit_condition(self, condition: Condition) -> typing.Optional[ConditionType]:
        return condition

    def visit_condition_group(
        self, condition_group: ConditionGroup
    ) -> typing.Optional[ConditionType]:
        conditions = []
        for condition in condition_group.conditions:
            visited = self.visit(condition)
            if visited:
                conditions.append(visited)

        condition_group.conditions = conditions

        return condition_group



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

from ..command import RobotoCommandSet
from .delete_image import delete_image_command
from .delete_repo import delete_repo_command
from .list_images import ls_images_command
from .list_repos import ls_repos_command
from .login import login_command
from .pull import pull_command
from .push import push_command

commands = [
    pull_command,
    push_command,
    ls_images_command,
    ls_repos_command,
    delete_image_command,
    delete_repo_command,
    login_command,
]

command_set = RobotoCommandSet(
    name="images",
    help=("Push and pull images from Roboto's private container registry."),
    commands=commands,
)



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

from .fga import (
    AuthZTupleRecord,
    EditAccessRequest,
    GetAccessResponse,
)
from .permissions import Permissions

__all__ = (
    "Permissions",
    "AuthZTupleRecord",
    "EditAccessRequest",
    "GetAccessResponse",
)



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import typing
import urllib.parse


class UrlBuilder:
    __endpoint: str

    def __init__(self, endpoint: str):
        self.__endpoint = endpoint.rstrip("/")

    def build(self, path: str, query: typing.Optional[dict[str, typing.Any]] = None):
        return UrlBuilder.url(endpoint=self.__endpoint, path=path, query=query)

    @staticmethod
    def url(
        endpoint: str, path: str, query: typing.Optional[dict[str, typing.Any]] = None
    ) -> str:
        normalized_path = path.lstrip("/")

        if query is None:
            return f"{endpoint}/{normalized_path}"
        else:
            return f"{endpoint}/{normalized_path}?{urllib.parse.urlencode(query)}"



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import datetime
from typing import Optional

import pydantic


class TokenContext(pydantic.BaseModel):
    """Context for token usage"""

    token_id: str
    name: str
    description: Optional[str] = None
    expires: datetime.datetime
    last_used: Optional[datetime.datetime] = None
    enabled: bool = True


class TokenRecord(pydantic.BaseModel):
    """A wire-transmissible representation of a token."""

    secret: Optional[str] = None
    user_id: Optional[str] = None
    context: Optional[TokenContext] = None



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import argparse

from ...domain.collections import Collection
from ..command import RobotoCommand
from ..common_args import add_org_arg
from ..context import CLIContext


def list(args, context: CLIContext, parser: argparse.ArgumentParser):
    for collection in Collection.list_all(
        owner_org_id=args.org,
        roboto_client=context.roboto_client,
    ):
        print(collection.record.model_dump_json(indent=2))


list_command = RobotoCommand(
    name="list",
    logic=list,
    setup_parser=add_org_arg,
    command_kwargs={"help": "Lists existing collections."},
)
